"""Core functionality"""

# from sklearn.cluster import KMeans
# from functools import partial

from warnings import warn
from functools import cached_property
from typing import Callable, Any, Optional, Mapping
from slang.stypes import (
    Waveform,
    Chunk,
    Chunker,
    Featurizer,
    Quantizer,
    Snip,
    Snips,
    FVs,
)
from slang.chunkers import DFLT_CHUNKER
from slang.featurizers import DFLT_FEATURIZER, DFLT_QUANTIZER

WfCallback = Optional[Callable[[Waveform], Any]]


class TagWfStore:
    def __init__(self, wf_store, key_to_tag=None, key_filt=None):
        self.wf_store = wf_store
        self.key_to_tag = key_to_tag
        self.key_filt = key_filt

    def key_wf_gen(self):
        pass

    def wfs(self):
        pass

    def tag_wfs(self):
        pass


class WfSource:
    def __init__(self, wfs, key_filt=None):
        self.wfs = wfs
        self.key_filt = key_filt

    @cached_property
    def keys(self):
        return tuple(filter(self.key_filt, self.wfs))

    def wf_gen(self):
        for k in self.keys:
            yield self.wfs[k]

    def key_wf_gen(self):
        for k in self.keys:
            yield k, self.wfs[k]


class AnnotedWfSource(WfSource):
    def __init__(self, wfs, annots=None, key_filt=None):
        super().__init__(wfs, key_filt)
        self.annots = annots

    @cached_property
    def keys(self):
        annots_keys = set(self.annots)
        return tuple([k for k in super().keys() if k in annots_keys])

    def annots_gen(self):
        for k in self.keys:
            yield self.annots[k]

    def annot_wf_gen(self):
        for k in self.keys:
            yield self.annots[k], self.wfs[k]

    def key_annot_wf_gen(self):
        for k in self.keys:
            yield self.annots[k], self.wfs[k]


# Notes:
"""
- key_to_tag not general enough because tag not general enough. Sometimes it's not a categorical.
    Sometimes it's multiple. Sometimes we have context information that needs to be associated with the annot. 
"""


class KvDataSource:
    def __init__(self, kv_store, key_to_tag=None, key_filt=None):
        self.kv_store = kv_store
        self.wfs = kv_store  # normalization alias
        self.key_to_tag = key_to_tag
        self.key_filt = key_filt

    def key_wf_gen(self):
        kv_store = self.kv_store
        for k in filter(self.key_filt, self.kv_store):
            yield k, kv_store[k]

    def key_tag_wf_gen(
        self, wf_callback: WfCallback = None, iterate_over_wf_callaback_output=False
    ):
        assert (
            self.key_to_tag is not None
        ), 'You need to have a key_to_tag function to do that!'
        if wf_callback is None:
            for k, wf in self.key_wf_gen():
                yield k, self.key_to_tag(k), wf
        else:
            assert callable(wf_callback), 'wf_callback needs to be callable.'

            if iterate_over_wf_callaback_output:
                for k, wf in self.key_wf_gen():
                    tag = self.key_to_tag(k)
                    for x in wf_callback(wf):
                        yield k, tag, x
            else:
                for k, wf in self.key_wf_gen():
                    yield k, self.key_to_tag(k), wf_callback(wf)

    def key_tag_chks_gen(self, wf_to_chk):
        yield from self.key_tag_wf_gen(
            wf_callback=wf_to_chk, iterate_over_wf_callaback_output=True
        )

    def key_tag_fvs_gen(self, wf_to_chk, chk_to_fv):
        wf_to_fv = lambda wf: list(map(chk_to_fv, wf_to_chk(wf)))
        yield from self.key_tag_wf_gen(
            wf_callback=wf_to_fv, iterate_over_wf_callaback_output=True
        )

    def key_tag_snips_gen(self, wf_to_chk, chk_to_fv, fv_to_snip):
        def wf_to_snips(wf):
            return list(fv_to_snip(chk_to_fv(chk)) for chk in wf_to_chk(wf))

        yield from self.key_tag_wf_gen(
            wf_callback=wf_to_snips, iterate_over_wf_callaback_output=True
        )

    def key_chks_gen(self, wf_to_chk):
        for k, wf in self.key_wf_gen():
            for chk in wf_to_chk(wf):
                yield k, chk

    def key_fvs_gen(self, wf_to_chk, chk_to_fv):
        for k, chk in self.key_chks_gen(wf_to_chk):
            yield k, chk_to_fv(chk)

    def key_snips_gen(self, wf_to_chk, chk_to_fv, fv_to_snip):
        for k, fv in self.key_fvs_gen(wf_to_chk, chk_to_fv):
            yield k, fv_to_snip(fv)

    def chk_tag_pairs(self, wf_to_chks):
        return ((chk, tag) for _, tag, chk in self.key_tag_chks_gen(wf_to_chks))

    def fv_tag_pairs(self, wf_to_chks, chk_to_fv):
        return ((fv, tag) for _, tag, fv in self.key_tag_fvs_gen(wf_to_chks, chk_to_fv))


class Snipper:
    """A base class that implements the wf->chk->fv->snip pipeline.
    Default functions for wf_to_chk (a.k.a. chunker), chk_to_fv (a.k.a. featurizer) and fv_to_snip (a.k.a. nipper)
    are given, but usually the user should specify these, and usually these are learned from data.
    """

    def __init__(
        self,
        wf_to_chks: Chunker = DFLT_CHUNKER,
        chk_to_fv: Featurizer = DFLT_FEATURIZER,
        fv_to_snip: Quantizer = DFLT_QUANTIZER,
    ):
        self.wf_to_chks = wf_to_chks
        self.chk_to_fv = chk_to_fv
        self.fv_to_snip = fv_to_snip

    def wf_to_fvs(self, wf: Waveform) -> FVs:
        for chk in self.wf_to_chks(wf):
            yield self.chk_to_fv(chk)

    def chk_to_snip(self, chk: Chunk) -> Snip:
        return self.fv_to_snip(self.chk_to_fv(chk))

    def wf_to_snips(self, wf: Waveform) -> Snips:
        for chk in self.wf_to_chks(wf):
            fv = self.chk_to_fv(chk)
            yield self.fv_to_snip(fv)

    snips_of_wf = wf_to_snips  # alias for back-compatibility

    # Delegations ##################

    # TODO: Get delegated attrs to show up as full fledged attr (tab complete, etc.)
    _delegations = dict(
        stats_of_snip=('fv_to_snip', 'stats_of_snip'),
        fv_of_snip=('fv_to_snip', 'fv_of_snip'),
        alphabet_size=('fv_to_snip', 'alphabet_size'),
    )

    def __getattr__(self, attr):
        """Delegate method to wrapped store if not part of wrapper store methods"""
        attr_spec = self._delegations.get(attr, None)
        if attr_spec is not None:
            a = self
            for aa in attr_spec:
                a = getattr(a, aa)
            return a
        else:
            raise AttributeError(f'Unknown attribute: {attr}')

    # TODO: Revise approach here (see above)
    #  lazyprop or property?
    #  getattr or self.attr (with try/catch or not)
    #  self.metric[snip] or self.all_metrics[snip][metric] or self.all_metrics[metric][snip] or self.all.metric[snip]
    # TODO: Delegate multiple at once?
    # @lazyprop
    # def stats_of_snip(self) -> Mapping:
    #     """(Lazy) property that holds the dict of snip stats"""
    #     return getattr(self.fv_to_snip, 'stats_of_snip', dict())  # look for them in fv_to_snip or return empty dict
    #
    # @property
    # def fv_of_snip(self) -> Mapping:
    #     """property that holds the dict of snip stats"""
    #     return getattr(self.fv_to_snip, 'fv_of_snip', dict())  # look for them in fv_to_snip or return empty dict

    @property
    def alphabet_size(self) -> int:
        return self.fv_to_snip.alphabet_size

    def __call__(self, wf: Waveform) -> Snips:
        return self.wf_to_snips(wf)



"""Types"""

from typing import Callable, Union, Any, List, Tuple
from numpy import ndarray, int16, int32, float32, float64

# Note I'm using Tuple to denote fixed sized sequences and List to denote a sliceable unbounded sized iterator and
# tuple as a fixed

from typing import TypeVar, Optional, Iterable, Sequence
from numbers import Number

from atypes import MyType

from atypes import (
    FixedSizeSeq,
    VarSizeSeq,
    Key,
    Sample,
    Waveform,
    Waveforms,
    KeyWfGen,
    Chunk,
    Chunks,
    Chunker,
    Feature,
    FV,
    FVs,
    Featurizer,
    ChkFeaturizer,
    Snip,
    Snips,
    Quantizer,
    Snipper,
)



"""Signal transformers
Tools to transform signals into a structured language.
"""
from slang.core import KvDataSource, Snipper

from slang.util import snips_to_str as dflt_snips_to_str
from slang.util import ddir

from slang.chunkers import (
    fixed_step_chunker,
    DFLT_CHUNKER,
    DFLT_CHK_SIZE,
    mk_chunker,
    fannout_chunker,
)

from slang.featurizers import (
    DFLT_WF_TO_SPECTR,
    DFLT_FEATURIZER,
    DFLT_QUANTIZER,
    tile_fft,
    mk_wf_to_spectr,
)

from slang.spectrop import (
    SpectralProjector,
    mk_chk_fft,
    geo_mat_11,
    logarithmic_bands_matrix,
)

from slang.snippers import (
    DfltWfToChk,
    PcaChkToFv,
    DfltChkToFv,
    FittableSnipper,
    ClassificationSnipper,
    FvToSnip,
    KMeansFvToSnip,
    KMeansFvToSnipDist,
    DfltFvToSnip,
)



"""Featurizers. Functions to get from waveform to feature vectors"""

from numpy import hanning, kaiser, abs, diff, array, std
from numpy.fft import rfft
from functools import wraps, partial
from typing import Sized, Optional, Callable
from operator import mul

from dol import Pipe
from slang.stypes import Chunk, Chunks, Featurizer


########################################################################################################################
# FFT

DFLT_WIN_FUNC = hanning


def tile_fft(tile, window=DFLT_WIN_FUNC, amp_function=abs):
    """Compute the power fft for a single tile
    """
    if callable(window):
        window = window(len(tile))
    fft_amplitudes = amp_function(rfft(tile * window))
    return fft_amplitudes


def identity_func(x):
    """The identify (a.k.a. transparent) function that returns it's input as is."""
    return x


def mk_window_func(window_func, *args, **kwargs):
    """Make a 'window function', which transforms input arrays through an elementsize
    multiplication by a fixed vector"""
    return partial(mul, window_func(*args, **kwargs))


mk_window_func.hanning = wraps(hanning)(partial(mk_window_func, hanning))
mk_window_func.kaiser = wraps(kaiser)(partial(mk_window_func, kaiser))


def _assert_size(a: Sized, size: Optional[int] = None, error_type=ValueError):
    if size is not None and len(a) != size:
        raise error_type(f'The size was expected to be {size} but was {len(a)}')
    return a


# https://github.com/otosense/slang/blob/2e6ccd208a9f1be0e1d46d134261c2e7807e84dc/slang/featurizers.py#L49


def mk_wf_to_spectr(
    preproc: Callable = None,
    fft_func: Callable = rfft,
    postproc: Callable = abs,
    *,
    assert_size: Optional[int] = None,
):
    """Make a function that computes the spectrogram of a waveform
    By spectrum, we mean the output of the pipeline:
        `tile -> preproc -> [assert_size] -> fft -> postproc -> spectrum

    Because typically, we preprocess the input waveform
    (say, transform with a hanning function), and post process
    the fft (say take the norm of the complex vector).

    Adds a

    :param preproc: Function to preprocess the array before computing the fft
    :param fft_func: Function that computes the fft
    :param postproc: Function post processing the fft array
    :param assert_size: If an int is given, the returned ``wf_to_spectr`` function
    will assert that the input array (or ``preproc(array)`` if ``preproc`` given) is of
    that size before computing the fft.

    :return: A ``wf_to_spectr`` function that computes the fft of a waveform (usually
    a chunk of fixed size).

    >>> import numpy as np
    >>> chk = np.ones(2048)  # the really interesting waveform we'll use as an example.
    >>> chk_size = len(chk)
    >>>
    >>> wf_to_spectr = mk_wf_to_spectr()  # default preproc is None (i.e. the waveform is not preprocessed)
    >>> s = wf_to_spectr(chk)
    >>> len(s)
    1025
    >>> assert s[1] == 0  # the second value is zero (with the hanning window, we wouldn't have that!)
    >>>
    >>> wf_to_spectr = mk_wf_to_spectr.w_hanning(chk_size)  # let's try the hanning window
    >>> s = wf_to_spectr(chk)
    >>> len(s)
    1025
    >>> assert s[1] > 0  # the second value is non-zero (without the hanning window, we wouldn't have that!)
    >>>
    >>> wf_to_spectr = mk_wf_to_spectr.w_kaiser(chk_size, beta=0.0)  # use kaiser preproc
    >>> s = wf_to_spectr(chk)
    >>> len(s)
    1025
    >>> assert s[1] == 0  # the second value is zero (because beta=0.0: with the hanning window, we wouldn't have that!)


    """
    if callable(preproc):
        wf_to_spectr = Pipe(preproc, fft_func, postproc)
    else:
        wf_to_spectr = Pipe(fft_func, postproc)
    wf_to_spectr.assert_size = assert_size
    return wf_to_spectr


def _mk_wf_to_spectr_w_hanning(
    window_size: int, fft_func: Callable = rfft, postproc: Callable = abs
):
    """Make a wf_to_spectr function that uses a hanning window preproc.
    """
    preproc = mk_window_func.hanning(window_size)
    return mk_wf_to_spectr(preproc, fft_func, postproc)


def _mk_wf_to_spectr_w_kaiser(
    window_size: int,
    beta: float = 0.0,
    fft_func: Callable = rfft,
    postproc: Callable = abs,
):
    """Make a wf_to_spectr function that uses a kaiser window preproc.

    The window_size should be set to the fixed tile (chunk) size you're using.
    The beta size is set to 0.0. To know what to set, look it up. We're using numpy's kaiser function for that.

    Here are some tips:

    ====  =======================
    beta  Window shape
    ====  =======================
    0     Rectangular
    5     Similar to a Hamming
    6     Similar to a Hanning
    8.6   Similar to a Blackman
    ====  =======================

    """
    preproc = mk_window_func.kaiser(window_size, beta)
    return mk_wf_to_spectr(preproc, fft_func, postproc)


mk_wf_to_spectr.w_hanning = _mk_wf_to_spectr_w_hanning
mk_wf_to_spectr.w_kaiser = _mk_wf_to_spectr_w_kaiser

DFLT_WF_TO_SPECTR = mk_wf_to_spectr()

########################################################################################################################
# Spectral Projectors
from numpy import dot

mat_mult = dot


class NotFittedError(ValueError, AttributeError):
    ...


# Featurizer
class SpectralProjector:
    def __init__(self, scalings_, chk_to_spectr=DFLT_WF_TO_SPECTR):
        self.scalings_ = scalings_
        self.chk_to_spectr = chk_to_spectr

    @property
    def fv_length(self):
        return self.scalings_.shape[1]

    def spectr_mat(self, chks: Chunks):
        return array(list(self.chk_to_spectr(chk) for chk in chks))

    def transform(self, chks: Chunks):
        return mat_mult(self.spectr_mat(chks), self.scalings_)

    def fv_of_chk(self, chk):
        return mat_mult(self.chk_to_spectr(chk), self.scalings_)

    def __call__(self, chk: Chunk):
        return list(self.chk_to_spectr(chk))


########################################################################################################################
# Some silly featurizer to play with

from bisect import bisect_left


def zcr(chk):
    return sum(diff(array(chk) > 0).astype(int)) / (len(chk) - 1)


def rms_zcr(chk):
    return std(chk), zcr(chk)


_n_levels = 10
_zcr_level_dividers = [2 ** (-(x + 1)) for x in range(0, _n_levels - 1)][::-1]
_rms_level_dividers = [2048 * 2 ** (-(x + 1)) for x in range(0, _n_levels - 1)][::-1]


def _rms_zcr_to_levels(rms, zcr):
    return (
        bisect_left(_rms_level_dividers, rms),
        bisect_left(_zcr_level_dividers, zcr),
    )


def rms_zcr_quantizer(fv):
    rms, zcr = fv
    rms_level, zero_crossing_level = _rms_zcr_to_levels(rms, zcr)
    return rms_level * _n_levels + zero_crossing_level


DFLT_FEATURIZER = rms_zcr
DFLT_QUANTIZER = rms_zcr_quantizer

import numpy as np
from typing import Union, Iterable


def mk_spectral_moment_featurizer(
    n_moments=100,
    preproc: Callable = None,
    fft_func: Callable = rfft,
    postproc: Callable = abs,
):
    wf_to_spectr = mk_wf_to_spectr(preproc, fft_func, postproc)
    moments = np.arange(1, n_moments + 1)
    n_moments = len(moments)
    std_exponents = np.arange(1, n_moments + 1)

    if n_moments > 2:

        def moment_featurizer(wf):
            a = wf_to_spectr(wf)
            a_std = np.std(a)
            m = moment(a, moments) / (a_std ** std_exponents)
            m[0] = np.mean(a)
            m[1] = a_std
            return m

    elif n_moments == 2:

        def moment_featurizer(wf):
            a = wf_to_spectr(wf)
            return np.array([np.mean(a), np.std(a)])

    elif n_moments == 1:

        def moment_featurizer(wf):
            a = wf_to_spectr(wf)
            return np.array([np.mean(a)])

    else:
        raise ValueError(
            f'n_moments should be a positive integer. Instead, was {n_moments}'
        )

    return moment_featurizer


def _chk_asarray(a, axis):
    if axis is None:
        a = np.ravel(a)
        outaxis = 0
    else:
        a = np.asarray(a)
        outaxis = axis

    if a.ndim == 0:
        a = np.atleast_1d(a)

    return a, outaxis


def moment(a, moment: Union[int, Iterable] = 1, axis=0):
    r"""Calculate the nth moment about the mean for a sample.
    Taken from scipy.stats

    Examples
    --------
    >>> moment([1, 2, 3, 4, 5], moment=1)
    0.0
    >>> moment([1, 2, 3, 4, 5], moment=2)
    2.0
    """
    a, axis = _chk_asarray(a, axis)

    if a.size == 0:
        # empty array, return nan(s) with shape matching `moment`
        if np.isscalar(moment):
            return np.nan
        else:
            return np.full(np.asarray(moment).shape, np.nan, dtype=np.float64)

    # for array_like moment input, return a value for each.
    if not np.isscalar(moment):
        mmnt = [_moment(a, i, axis) for i in moment]
        return np.array(mmnt)
    else:
        return _moment(a, moment, axis)


def _moment(a, moment, axis):
    if np.abs(moment - np.round(moment)) > 0:
        raise ValueError('All moment parameters must be integers')

    if moment == 0:
        # When moment equals 0, the result is 1, by definition.
        shape = list(a.shape)
        del shape[axis]
        if shape:
            # return an actual array of the appropriate shape
            return np.ones(shape, dtype=float)
        else:
            # the input was 1D, so return a scalar instead of a rank-0 array
            return 1.0

    elif moment == 1:
        # By definition the first moment about the mean is 0.
        shape = list(a.shape)
        del shape[axis]
        if shape:
            # return an actual array of the appropriate shape
            return np.zeros(shape, dtype=float)
        else:
            # the input was 1D, so return a scalar instead of a rank-0 array
            return np.float64(0.0)
    else:
        # Exponentiation by squares: form exponent sequence
        n_list = [moment]
        current_n = moment
        while current_n > 2:
            if current_n % 2:
                current_n = (current_n - 1) / 2
            else:
                current_n /= 2
            n_list.append(current_n)

        # Starting point for exponentiation by squares
        a_zero_mean = a - np.expand_dims(np.mean(a, axis), axis)
        if n_list[-1] == 1:
            s = a_zero_mean.copy()
        else:
            s = a_zero_mean ** 2

        # Perform multiplications
        for n in n_list[-2::-1]:
            s = s ** 2
            if n % 2:
                s *= a_zero_mean
        return np.mean(s, axis)



"""Spectral projectors"""

from functools import wraps, partial
import inspect
from dataclasses import dataclass, field
from typing import Union, Optional, Callable
import warnings
from math import sqrt

from numpy import cumsum, min, inf, ndarray, floor
from sklearn.decomposition import PCA, IncrementalPCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.utils.validation import NotFittedError
from sklearn.base import BaseEstimator, TransformerMixin

import numpy as np
from numpy import (
    array,
    hanning,
    fft,
)  # TODO: Get rid of this once we use C-based spectr

DFLT_WIN_FUNC = hanning
DFLT_CHK_SIZE = 2048
DFLT_INPUT_SIZE = int(1 + DFLT_CHK_SIZE / 2)  # 1025 for chk sie of 2048
DFLT_N_FREQ = DFLT_INPUT_SIZE  # backcompat alias
DFLT_AMPLITUDE_FUNC = np.abs
DFLT_FREQ_RANGE = (0, DFLT_N_FREQ)


def make_band_matrix_row(list_entries, row_len):
    """
    Makes a row for the spectral bucket matrix. The row is zero everywhere except on the entries of index in
    list_entries where it is 1 / len(list_entries)

    :param list_entries: the indices of non zero entries of the row
    :param row_len: the length of the row
    :return: an array of length row_len as described above

    >>> make_band_matrix_row([3], 5)
    array([0., 0., 0., 1., 0.])
    >>> make_band_matrix_row([0, 3], 5)
    array([0.5, 0. , 0. , 0.5, 0. ])

    """

    n_non_zero = len(list_entries)
    row = np.array(
        [0 if i not in list_entries else 1 / n_non_zero for i in range(row_len)]
    )
    return row


def make_band_matrix(buckets, n_freq=DFLT_INPUT_SIZE):
    """
    Given a list of n list of indices, make a matrix of size n by n_freq where the entries of row k are 0
    everywhere except at the index in the k list of buckets, where the entries are the inverse of the length
    of that bucket

    :param buckets: a list of list containing the indices of non zero entries of the corresponding row
    :param n_freq: the number of column of the matrix
    :return: a len(buckets) by n_freq matrix

    >>> buckets = make_buckets(n_buckets=15, freqs_weighting=lambda x: np.log(x + 0.001), freq_range=(200, 1000))
    >>> M = make_band_matrix(buckets, n_freq=1025)
    >>> print(M.shape)
    (15, 1025)

    >>> # the matrix sends each of the 200 first unitary vector to the zero vector (not below is not a mathematical proof of that, but solid clue)
    >>> vec = [1] * 200 + [0] * (1025 - 200)
    >>> print(np.dot(M, vec))
    [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]

    >>> # after that, we get non zero output (in general at least)
    >>> vec = [1] * 200 + [1] + [0] * (1025 - 201)
    >>> print(np.dot(M, vec))
    [0.01612903 0.         0.         0.         0.         0.
     0.         0.         0.         0.         0.         0.
     0.         0.         0.        ]

    """

    n_bands = len(buckets)
    bucket_matrix = np.zeros((n_bands, n_freq))
    for row_idx, bucket in enumerate(buckets):
        row = make_band_matrix_row(list_entries=bucket, row_len=n_freq)
        bucket_matrix[row_idx] = row
    return bucket_matrix


def decreasing_integer_geometric_sequence(
    start: int = DFLT_INPUT_SIZE, scale_factor=0.5
) -> list:
    """Generate decreasing positive integers in by multiplying numbers by a constant (between 0 and 1) repeatedly.
    Numbers of ther sequence will all be integers and not repeat (so often not a true geometri sequence).
    All sequences will end with 1.
    An error will be raised if scale_factor is not between 0 and 1 (exclusive).

    >>> decreasing_integer_geometric_sequence(128)
    [128, 64, 32, 16, 8, 4, 2, 1]
    >>> decreasing_integer_geometric_sequence(10, 0.3)
    [10, 3, 1]
    >>> decreasing_integer_geometric_sequence(10, 0.7)
    [10, 7, 5, 3, 2, 1]

    And to see that we indeed don't get duplicates of the true geometric sequence.

    >>> decreasing_integer_geometric_sequence(10, 0.9999)
    [10, 9, 8, 7, 6, 5, 4, 3, 2, 1]

    """
    assert (
        0 < scale_factor < 1
    ), 'This geometric_sequence is meant for decreasing sequences only'

    def gen():
        cursor = start
        yield round(cursor)
        while cursor > 1:
            cursor *= scale_factor
            yield round(cursor)

    return list(dict.fromkeys(gen()))


# TODO: Continue algorithm further, filling more coarse coverage with static greedy
#  rule, thus making n_buckets unbound
def logarithmic_bands_matrix(
    n_buckets: Optional[int] = None,
    n_freqs: int = DFLT_INPUT_SIZE,
    factor: Optional[float] = None,
) -> np.ndarray:
    """Makes a spectral projection matrix that puts more importance on low frequencies than high ones.
    Importance both in weight and in precision.
    By a factor of 2 by default, but can be any amount.
    Note that the factor here is the inverse of the geometric factor of decreasing_integer_geometric_sequence.

    Note that the sum of the frequencies of the bands is constant from band to band. Not sure if that's the best choice,
    or if equal weight through out is the best choice.
    We can discuss what both mean at a later date.

    Important here is to understand my intent, to see how well I achieve it.
    - The features "build on each other". That is, if you ask (within the same
    `(n_freqs, factor)` set) for 7 features in your fv, the first 5 will be the same
    as if you asked for only 5. This makes it easier to compare
    fvs between projects, and possibly even "add to existing features".
    - Note  that the first feature is always total energy.
    - Lower frequencies are given less importance -- both in precision and in weight.
    - The linear algebraists that are listening will note what the vector space
    generated actually is, compared to that created by disjoint bands.

    >>> print(*logarithmic_bands_matrix(n_freqs=8).tolist(), sep='\\n')
    [0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125]
    [0.25, 0.25, 0.25, 0.25, 0.0, 0.0, 0.0, 0.0]
    [0.5, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    >>> print(*logarithmic_bands_matrix(n_freqs=8, factor=1.5).tolist(), sep='\\n')
    [0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125]
    [0.2, 0.2, 0.2, 0.2, 0.2, 0.0, 0.0, 0.0]
    [0.25, 0.25, 0.25, 0.25, 0.0, 0.0, 0.0, 0.0]
    [0.5, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    """
    if n_buckets is None and factor is None:
        factor = 2
    bucket_upper_bounds = _compute_bucket_upper_bounds(n_buckets, n_freqs, factor)

    buckets = list(range(0, i) for i in bucket_upper_bounds)
    m = make_band_matrix(buckets, n_freq=n_freqs)
    if n_buckets is not None:
        assert n_buckets <= len(m), (
            f'you asked for {n_buckets}, but the matrix has only {len(m)} rows. '
            f'Consider decreasing the factor to get more buckets'
        )
    else:
        n_buckets = len(m)
    return m[:n_buckets]


def _compute_bucket_upper_bounds(n_buckets, n_freqs, factor, max_tries=20):
    """Computes bucket upper bounds, trying to compensate for float numerical errors.

    If factor is None, the formula to get the smallest factor that is needded to get
    the given ``n_buckets`` is ``factor = n_freqs ** (1 / (n_buckets - 1))``.
    Unfortunately, this theoretical truth doesn't reflect in physical computation
    due to float precision errors. Our hack here is to ask for the factor for a higher
    number of buckets until we get a factor that is "sufficient".
    """
    if factor is None:
        # TODO: This formula is not bullet proof. We tried to compensate for numerical
        #  errors for reasonably small n_buckets and n_freqs
        for excess in range(max_tries):
            factor = n_freqs ** (1 / (n_buckets + excess - 1))
            bucket_upper_bounds = decreasing_integer_geometric_sequence(
                n_freqs, 1 / factor
            )
            if len(bucket_upper_bounds) > n_buckets:
                return bucket_upper_bounds
        if excess == max_tries - 1:
            raise ValueError(f'Maximum number of tries exceeded')

    else:
        return decreasing_integer_geometric_sequence(n_freqs, 1 / factor)


# A few default spectral projection matrices
geo_mat_11 = logarithmic_bands_matrix(11, n_freqs=1025, factor=2)
geo_mat_17 = logarithmic_bands_matrix(17, n_freqs=1025, factor=1.5)
geo_mat_34 = logarithmic_bands_matrix(34, n_freqs=1025, factor=1.2)

DFLT_SCALINGS = geo_mat_11


def instantiate_class_and_inject_attributes(cls, **kwargs):
    """
    instantiates a class with the given kwargs, picking those arguments that are in the signature of cls to
    use for the __init__, and adding attributes to the constructed object with the remaining.
    :param cls: class to insantiate
    :param kwargs: keyword args (some for the class __init__, and others to inject)
    :return: An (possibly enhanced) class instance
    >>> class C:
    ...     def __init__(self, a, b=3):
    ...         self.a = a
    ...         self.b = b
    ...
    >>> c = instantiate_class_and_inject_attributes(C, a=10, foo='bar')
    >>> c.__dict__
    {'a': 10, 'b': 3, 'foo': 'bar'}
    >>> c = instantiate_class_and_inject_attributes(C, a=10, foo='bar', bar='foo', b=1000)
    >>> c.__dict__
    {'a': 10, 'b': 1000, 'foo': 'bar', 'bar': 'foo'}
    >>> try:
    ...     c = instantiate_class_and_inject_attributes(C, foo='bar', bar='foo', b=1000)
    ... except TypeError:  # expected
    ...     pass
    """
    # cls_signature_args = inspect.signature(cls).parameters
    cls_signature_args = inspect.getfullargspec(cls.__init__).args[1:]

    cls_kwargs = dict()
    other_kwargs = dict()
    for k, v in kwargs.items():
        if k in cls_signature_args:
            cls_kwargs[k] = v
        else:
            other_kwargs[k] = v
    o = cls(**cls_kwargs)
    o.__dict__.update(other_kwargs)
    return o


def ascertain_array(iterable):
    """Getting an array from an iterable"""
    if iterable is not None and not isinstance(iterable, ndarray):
        if not isinstance(iterable, list):
            iterable = list(iterable)
        iterable = array(iterable)
    return iterable


def chk_to_spectrum(
    chk, chk_size, window=DFLT_WIN_FUNC, amplitude_func=DFLT_AMPLITUDE_FUNC
):
    assert len(chk) == chk_size, (
        f'This function was made for chk_size={chk_size}. '
        f'You fed a chk of size len(chk)={len(chk)} instead'
    )
    fft_amplitudes = amplitude_func(np.fft.rfft(chk * window))
    return fft_amplitudes


def named_partial(name, func, *args, **kwargs):
    partial_func = partial(func, *args, **kwargs)
    partial_func.name = name
    return partial_func


def mk_chk_fft(chk_size=None, window=DFLT_WIN_FUNC, amplitude_func=DFLT_AMPLITUDE_FUNC):
    """Make a chk_fft function that will compute the fft (with given amplitude and window).
    Note that this output chk_fft function will enforce a fixed chk_size (given explicitly, or through the
    size of the window (if window is given as an array)

    >>> chk_size = 4 * 5
    >>> f = mk_chk_fft(chk_size)
    >>> chk = 4 * list(range(chk_size // 4))
    >>> f(chk)
    array([19.        , 10.214421  ,  0.40774689,  4.34150779,  8.09801978,
            4.30684381,  0.26711259,  2.61514496,  5.04588504,  2.6255033 ,
            0.21962565])
    >>> # verifying that it's pickable
    >>> import pickle
    >>> import numpy as np
    >>> ff = pickle.loads(pickle.dumps(f))
    >>> assert np.allclose(ff(chk), ff(chk))
    """
    if callable(window):
        assert (
            chk_size is not None
        ), 'chk_size must be a positive integer if window is a callable, or None'
        window = window(chk_size)
    elif window is None:
        window = 1
    else:
        window = np.array(window)
        if chk_size is not None:
            assert (
                len(window) == chk_size
            ), f"chk_size ({chk_size}) and len(window) ({len(window)}) don't match"

    chk_spectrum = named_partial(
        'chk_to_spectrum',
        chk_to_spectrum,
        chk_size=chk_size,
        window=window,
        amplitude_func=amplitude_func,
    )
    chk_spectrum.chk_size = chk_size
    return chk_spectrum


DFLT_CHK_FFT = mk_chk_fft(chk_size=DFLT_CHK_SIZE, window=DFLT_WIN_FUNC)

matrix_mult = np.dot
DFLT_MATRIX_MULTI = matrix_mult


# TODO: Does the doc match what the function does?
# TODO: Better name for this, to distinguish between expressing projection in projected sub-space or original space
#   The following expresses it in the original space
def projection(basis, vectors, mat_mult=DFLT_MATRIX_MULTI):
    """
    The vectors live in a k dimensional space S and the columns of the basis are vectors of the same
    space spanning a subspace of S. Gives a representation of the projection of vector into the space
    spanned by basis in term of the basis.

    :param basis: an n-by-k array, a matrix whose vertical columns are the vectors of the basis
    :param vectors: an m-by-k array, a vector to be represented in the basis
    :param mat_mult: the function to multiply matrices
    :return: an m-by-k array
    """
    return mat_mult(mat_mult(vectors, basis), basis.T)


def reducing_proj(basis, vectors, mat_mult=DFLT_MATRIX_MULTI):
    """What we actually use to get fvs from spectras"""
    return mat_mult(vectors, basis.T)


def _assert_spectrum_and_chk_size_match(spectrum_size: int, chk_size: int):
    assert spectrum_size == int(1 + chk_size / 2), (
        f'expected spectrum_size == int(1 + chk_size / 2): '
        f'{spectrum_size=} and {chk_size=}'
    )


def _chk_size_to_spectrum_size(chk_size: int) -> int:
    return int(1 + chk_size / 2)


def _spectrum_size_to_chk_size(spectrum_size: int) -> int:
    if spectrum_size == 1:
        return 1
    return 2 * spectrum_size - 2


def _get_spectrum_size_and_validate_against_chk_size(spectrum_size, chk_size):
    assert (
        chk_size is not None or spectrum_size is not None
    ), 'either chk_size or spectrum size must be given'
    if spectrum_size is None:
        if chk_size is not None:
            spectrum_size = int(1 + chk_size / 2)
        else:
            _assert_spectrum_and_chk_size_match(spectrum_size, chk_size)
    return spectrum_size


@dataclass
class Projector:
    scalings_: np.ndarray = field(default_factory=lambda: DFLT_SCALINGS)
    mat_mult: Callable = DFLT_MATRIX_MULTI

    def transform(self, X):
        # return projection(self.scalings_, ascertain_array(X))
        return reducing_proj(self.scalings_, ascertain_array(X), mat_mult=self.mat_mult)

    def to_jdict(self):
        return {
            'scalings_': self.scalings_.tolist(),
        }

    @classmethod
    def from_jdict(cls, jdict):
        obj = instantiate_class_and_inject_attributes(cls, **jdict)
        if hasattr(obj, 'scalings_'):
            obj.scalings_ = array(obj.scalings_)
        return obj

    def is_fitted(self):
        return hasattr(self, 'scalings_') and self.scalings_ is not None

    def assert_is_fitted(self):
        if not self.is_fitted():
            raise NotFittedError(
                '{} was not fitted yet.'.format(self.__class__.__name__)
            )

    def __call__(self, vector):
        return self.transform([vector])[0]


@dataclass
class SpectralProjector(Projector):
    chk_fft: Callable = DFLT_CHK_FFT

    def __post_init__(self):
        if not hasattr(self, 'chk_size'):
            self.chk_size = getattr(
                self.chk_fft, 'chk_size', None
            )  # get chk_size from chk_fft if it has it

    # TODO: spectrum_size is elsewhere n_freqs (should we be consistent?)
    @classmethod
    def for_sizes(
        cls,
        chk_size: Optional[int] = None,
        n_features: Optional[int] = None,
        *,
        spectrum_size: Optional[int] = None,
        log_factor: Optional[float] = None,
    ):
        """Makes a projector for given (chunk or spectrum) and feature space size.
        This is a convenience to get a projector given only the input
        (chunk or spectrum) size, and optionally output (feature space) size.
        That is, we only need minimal information about the data here, not to fit
        on the data itself.

        >>> featurizer = SpectralProjector.for_sizes(chk_size=20)
        >>> featurizer.scalings_.shape
        (4, 11)
        >>> SpectralProjector.for_sizes(chk_size=21).scalings_.shape
        (4, 11)
        >>> SpectralProjector.for_sizes(chk_size=21, n_features=2).scalings_.shape
        (2, 11)
        >>> SpectralProjector.for_sizes(spectrum_size=21).scalings_.shape
        (5, 21)

        >>> featurizer = SpectralProjector.for_sizes(chk_size=4)
        >>> chk = [1, 2, 3, 4]
        >>> fv = featurizer(chk).round(decimals=3)
        >>> fv
        array([2.401, 3.227])

        >>> spectrum = featurizer.spectras([chk])[0]
        >>> spectrum.round(decimals=2)
        array([3.75, 2.7 , 0.75])
        >>> featurizer.scalings_.round(2)
        array([[0.33, 0.33, 0.33],
               [0.5 , 0.5 , 0.  ]])
        >>> fv == featurizer.mat_mult(featurizer.scalings_, spectrum).round(decimals=3)
        array([ True,  True])

        """
        spectrum_size = _get_spectrum_size_and_validate_against_chk_size(
            spectrum_size=spectrum_size, chk_size=chk_size
        )
        # TODO: Research some scientifically based default n_features based on spectrum
        n_features = n_features or min((50, int(1 + sqrt(spectrum_size))))
        scalings_ = logarithmic_bands_matrix(
            n_buckets=n_features, n_freqs=spectrum_size, factor=log_factor
        )
        instance = cls(scalings_)
        if chk_size is not None:
            instance.chk_size = chk_size
        instance.spectrum_size = spectrum_size
        instance.chk_fft = mk_chk_fft(chk_size=instance.chk_size, window=DFLT_WIN_FUNC)
        return instance

    def spectras(self, chks):
        return array([self.chk_fft(chk) for chk in chks])

    def transform(self, chks):
        return super().transform(self.spectras(chks))

    def __call__(self, chk):
        return self.transform([chk])[0]

    mk_chk_fft = staticmethod(
        mk_chk_fft
    )  # to have it available to make a chk_fft in __init__


# @classmethod
# def for_chk_size(cls, scalings_=None,
#                  chk_size=None, window=DFLT_WIN_FUNC, amplitude_func=DFLT_AMPLITUDE_FUNC,
#                  mat_mult=DFLT_MATRIX_MULTI):
#     chk_fft = mk_chk_fft(chk_size, window, amplitude_func)
#     return cls(scalings_, chk_fft, mat_mult)


# def handle_iterables(learner_cls):
#     @wraps(learner_cls.fit)
#     def fit(self, X, y=None, *args, **kwargs):
#         # TODO use signature bind to extract the X and y from args, and kwargs (and get rid of X and y explicit arg)
#         try:
#             self.fit(X, y, *args, **kwargs)
#         except Exception:  # TODO: Be less broad
#             # Use chunking to fit partial in batches (probably more efficient)
#             if hasattr(self, 'fit_partial'):
#                 for x_item, y_item in zip(X, y):
#                     self.fit_partial(x_item, y_item)
#             else:
#                 X = ascertain_array(X)
#                 y = ascertain_array(y)
#                 self.fit(X, y, *args, **kwargs)
#         return self
#     learner_cls.fit = fit


def fit_handling_iterables(learner, X, y=None, *args, **kwargs):
    try:
        learner.fit(X, y, *args, **kwargs)
    except Exception:  # TODO: Be less broad
        # Use chunking to fit partial in batches (probably more efficient)
        if hasattr(learner, 'fit_partial'):
            for x_item, y_item in zip(X, y):
                learner.fit_partial(x_item, y_item)
        else:
            X = ascertain_array(X)
            y = ascertain_array(y)
            learner.fit(X, y, *args, **kwargs)
    return learner


class SpectralProjectorLearner(SpectralProjector, BaseEstimator, TransformerMixin):
    def __init__(self, learner, chk_fft=DFLT_CHK_FFT, mat_mult=DFLT_MATRIX_MULTI):
        super().__init__(scalings_=None, chk_fft=chk_fft, mat_mult=mat_mult)
        self.learner = learner

    def fit(self, chks, y=None):
        spectras = map(self.chk_fft, chks)
        fit_handling_iterables(self.learner, X=spectras, y=y)
        self.scalings_ = self.learner.scalings_
        return self


# TODO: Probably better if subclassing PCA
class TargettedVariancePCA(BaseEstimator, TransformerMixin):
    def __init__(self, target_variance=0.95, max_n_components=inf, min_n_components=1):
        self.target_variance = target_variance
        self.max_n_components = max_n_components
        self.min_n_components = min_n_components

    def fit(self, X, y=None):
        max_possible_comp = min(X.shape)
        self.max_n_components = int(min([self.max_n_components, max_possible_comp]))
        pca = PCA(n_components=self.max_n_components)
        pca.fit(X)
        exp_var_rat = pca.explained_variance_ratio_
        sum_exp_var_rat = cumsum(exp_var_rat)

        n_comp_to_retain = 0
        for n_comp_to_retain in range(self.min_n_components, self.max_n_components + 1):
            if sum_exp_var_rat[n_comp_to_retain - 1] >= self.target_variance:
                break

        pca.components_ = pca.components_[:n_comp_to_retain, :]

        # Note: pca not needed for the functioning of the class, but keeping around for debug reasons
        self._pca = pca

        self.scalings_ = pca.components_.T
        return self

    def transform(self, X):
        return matrix_mult(X, self.scalings_)


class SpectralProjectorUnsupervisedFitter(SpectralProjector, TargettedVariancePCA):
    def __init__(
        self,
        target_variance=0.95,
        max_n_components=inf,
        min_n_components=1,
        scalings_=None,
        chk_fft=DFLT_CHK_FFT,
    ):
        TargettedVariancePCA.__init__(
            self, target_variance, max_n_components, min_n_components
        )
        SpectralProjector.__init__(self, scalings_, chk_fft)

    def fit(self, chks, y=None):
        return super().fit(self.spectras(chks), y)


class MyLDA(BaseEstimator, TransformerMixin):
    def __init__(self, n_components=None):
        self.n_components = n_components

    def fit(self, X, y):
        lda = LinearDiscriminantAnalysis(n_components=self.n_components)
        lda.fit(X, ascertain_array(y))

        # Note: _lda not needed for the functioning of the class, but keeping around for debug reasons
        self._lda = lda

        self.scalings_ = lda.scalings_
        self.classes_ = lda.classes_
        return self

    def transform(self, X):
        return matrix_mult(X, self.scalings_)


class SpectralProjectorSupervisedFitter(SpectralProjector, MyLDA):
    def __init__(self, n_components=None, scalings_=None, chk_fft=DFLT_CHK_FFT):
        MyLDA.__init__(self, n_components)
        SpectralProjector.__init__(self, scalings_, chk_fft)

    def fit(self, chks, y):
        return super().fit(self.spectras(chks), y)


########################################################################################################################
# Utils to create and extend projections


import importlib
import numpy as np
from sklearn.neighbors import NeighborhoodComponentsAnalysis as NCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.decomposition import PCA
from sklearn.linear_model import LinearRegression

CLUSTERING_OPTIONS = (
    'KMeans',
    'SpectralClustering',
    'AffinityPropagation',
    'AgglomerativeClustering',
    'Birch',
    'MeanShift',
)


# Note: All these models (PCA, LDA, NCA, LinReg...) are learned with numpy.dot as matrix mult.
#   What happens if we learn with one definition of multiplication and run with another?

# TODO: Try changing all learn_spect_proj cases into subclasses of Projection and compare pros and cons.
# Note: Pro 1; Both cases and per-case arguments and defaults are visible, not hidden in code
# Note: Pro 2; Projection will wrap scalings_ and offer consistent methods for projection calculus
# Question: Do we want to keep scalings_ as the projecting matrix of Projection?
#   Both name and shape (transp) not consistent with pca, nca, linear regression, etc.

# TODO: Examples of objectivication below. Not verified or tested yet.
# Note: A choice to make here: Subclassing versus Delegation. I'll perform subclassing here.


class PcaProj(PCA, Projector):
    def fit(self, X, y=None):
        super().fit(X, y)
        self.scalings_ = self.components_
        return self

    transform = Projector.transform


class IpcaProj(IncrementalPCA, Projector):
    def fit(self, X, y=None):
        super().fit(X, y)
        self.scalings_ = self.components_
        return self

    transform = Projector.transform


class LdaProj(LinearDiscriminantAnalysis, Projector):
    transform = Projector.transform


class LinregProj(LinearRegression, Projector):
    @wraps(LinearRegression.fit)
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self.scalings_ = self.coef_.T
        return self


# Example of PCA+Projection, with delegation


class PcaProjWithDelegation(Projector):
    @wraps(PCA.__init__)
    def __init__(self, *args, **kwargs):
        self._source = PCA(*args, **kwargs)

    def fit(self, X, y=None):
        self._source.fit(X, y)
        self.scalings_ = self._source.components_
        return self

    transform = Projector.transform


def learn_spect_proj(
    X,
    y=None,
    spectral_proj_name='pca',
    clustering_meth='KMeans',
    clustering_options=CLUSTERING_OPTIONS,
    kwargs_feat=None,
    kwargs_clust=None,
):
    """
    Function to learn each of the important spectral projection

    :param X: the fvs, an array of size n*k
    :param y: the classes, an array of size n
    :param spectral_proj_name: a string of the name of the featurizer
    :param args: extra argument to be passed to the featurizer class
    :return: a matrix in the form of a numpy array
    """

    clustering_options = set(clustering_options)
    kwargs_feat = kwargs_feat or {'n_components': 10}
    kwargs_clust = kwargs_clust or {}

    assert (
        clustering_meth in clustering_options
    ), 'clustering options must one of {}'.format(
        ', '.join(map(str, clustering_options))
    )
    clusterer_m = getattr(importlib.import_module('sklearn.cluster'), clustering_meth)

    if spectral_proj_name == 'keep_features':
        indices = kwargs_feat['indices']
        proj_matrix = np.zeros((X.shape[1], len(indices)))
        for idx in range(len(indices)):
            proj_matrix[indices[idx], idx] = 1

    elif spectral_proj_name == 'pca':
        pca = PCA(**kwargs_feat)
        pca.fit(X)
        proj_matrix = pca.components_.T

    # elif spectral_proj_name == 'pseudo_pca':
    #     # make the pseudo pca proj matrix
    #     ppca = PseudoPca(**kwargs_feat)
    #     ppca.fit(X)
    #     proj_matrix = ppca.proj_mat.T

    elif spectral_proj_name == 'lda':
        lda = LDA(**kwargs_feat)
        lda.fit(X, y)
        n_components = kwargs_feat['n_components']
        proj_matrix = lda.scalings_[:, :n_components]

    elif spectral_proj_name == 'unsupervised_lda':
        n_components = kwargs_feat['n_components']
        if y is not None:
            print('y will be replaced by classes found by the chosen clusterer')
        if 'n_clusters' in clusterer_m.__init__.__code__.co_varnames:
            y = clusterer_m(n_clusters=n_components + 1, **kwargs_clust).fit_predict(X)
        else:
            y = clusterer_m(**kwargs_clust).fit_predict(X)
        lda = LDA(**kwargs_feat)
        lda.fit(X, y)
        proj_matrix = lda.scalings_[:, :n_components]

    elif spectral_proj_name == 'nca':
        nca = NCA(**kwargs_feat)
        nca.fit(X, y)
        proj_matrix = nca.components_.T

    elif spectral_proj_name == 'unsupervised_nca':
        n_components = kwargs_feat['n_components']
        if y is not None:
            print('y will be replaced by classes found by the chosen clusterer')
        if 'n_clusters' in clusterer_m.__init__.__code__.co_varnames:
            y = clusterer_m(n_clusters=n_components + 1, **kwargs_clust).fit_predict(X)
        else:
            y = clusterer_m(**kwargs_clust).fit_predict(X)
        nca = NCA(**kwargs_feat)
        nca.fit(X, y)
        proj_matrix = nca.components_.T

    elif spectral_proj_name == 'linear regression':
        lr = LinearRegression(**kwargs_feat)
        lr.fit(X, y)
        proj_matrix = lr.coef_.T

    else:
        all_spectral_proj = ', '.join(
            [
                'keep_features',
                'pca',
                'lda',
                'pseudo_pca',
                'unsupervised_lda',
                'unsupervised_nca',
                'nca',
                'linear regression',
            ]
        )
        raise ValueError(f'the spectral projector must be one of: {all_spectral_proj}')

    return proj_matrix


def residue(scalings, X):
    """
    find the residue of each of vectors after projection in basis

    residues will be vectors in the original space (same number of dimensions)

    :param scalings: an n-by-m array, spanning a vector space A
    :param X: an n-by-k array
    :return: an n-by-l array, the residues of vectors with the respect of the projection in basis

    >>> A = np.array([[1,0],[0,1]])
    >>> B = np.array([[2,3]])
    >>> print(residue(A, B))
    [[0 0]]
    >>> A = np.array([[1,0],[0,0]])
    >>> B = np.array([[2,3]])
    >>> print(residue(A, B))
    [[0 3]]
    """
    # assert X - projection(scalings, X) == X * (scalings - 1)
    return X - projection(scalings, X)


# TODO: Check. Not consistent with normal projection (scalings_ is not used as projection), though it is a linear trans.
# Note: Perhaps this is better seen as the compliment/dual of a projection?
class Residues(Projector):
    def transform(self, X):
        return X - projection(self.scalings_, X)


def keep_only_indices(indices, input_size=DFLT_INPUT_SIZE):
    """
    Makes a matrix which when a spectra is multiplied with it, only the entires in the list indices are kept
    :param indices: a list of indices to keep
    :param input_size: the total number of frequencies in the spectra
    :return: a matrix of size (input_size, len(indices))
    """

    proj_matrix = np.zeros((input_size, len(indices)))
    for idx in range(len(indices)):
        proj_matrix[indices[idx], idx] = 1
    return proj_matrix


def mk_pre_projection_from_indices(indices=None, input_size=DFLT_INPUT_SIZE):
    return keep_only_indices(indices, input_size=input_size)


def learn_chain_proj_matrix(
    X,
    y=None,
    chain=({'type': 'pca', 'args': {'n_components': 5}},),
    indices=None,
    input_size=1025,
):
    """
    A function successively learning a projections matrix on the residue of the previous one. The projections
    matrices are then concatenated and return as one single projection matrix. Note that the final projection
    matrix may not produce fvs of the size the sum of the components of each part, i.e., care must be taken
    to ensure each classes called must be able to produce the required number of components. For example,
    if the number of classes is 10, then lda can only produce 9 dimensions. To obtain say 12 dimension, the user
    will need to chain two lda's, for example with size 9 and 3 respectively.

    :param X: the fvs, an array of size n*k
    :param y: the classes, an array of size n
    :param chain: a tuple of dictionaries each containing the type of projection along with its parameters
    :param indices: the indices of the spectra to work with, anything else is discarded
    :param input_size: the total number of entries from the spectra.
        Only needed if n_freq is not None, in order to
        determine the size of the freq_selection_matrix
    :return: a single projection matrix
    """

    freq_selection_matrix = None
    if indices is not None:
        freq_selection_matrix = keep_only_indices(indices, input_size=input_size)
        X = np.dot(X, freq_selection_matrix)

    all_proj_matrices = []
    for mat_dict in chain:
        kwargs_feat = mat_dict['args']
        proj_matrix = learn_spect_proj(
            X, y, spectral_proj_name=mat_dict['type'], kwargs_feat=kwargs_feat
        )
        all_proj_matrices.append(proj_matrix)
        X = residue(proj_matrix, X)

    proj_matrix = np.hstack(tuple(all_proj_matrices))
    if indices is not None:
        proj_matrix = np.dot(freq_selection_matrix, proj_matrix)
    return np.array(proj_matrix)


def old_learn_chain_proj_matrix(
    X,
    y=None,
    chain=({'type': 'pca', 'kwargs': {'n_components': 5}},),
    indices=None,
    input_size=DFLT_INPUT_SIZE,
):
    freq_selection_matrix = None
    if indices is not None:
        freq_selection_matrix = keep_only_indices(indices, input_size=input_size)
        X = np.dot(X, freq_selection_matrix)

    all_proj_matrices = []
    for mat_dict in chain:
        kwargs_feat = mat_dict['kwargs']
        proj_matrix = learn_spect_proj(
            X, y, spectral_proj_name=mat_dict['type'], kwargs_feat=kwargs_feat
        )
        all_proj_matrices.append(proj_matrix)
        X = residue(proj_matrix, X)

    proj_matrix = np.hstack(tuple(all_proj_matrices))
    if indices is not None:
        proj_matrix = np.dot(freq_selection_matrix, proj_matrix)
    return np.array(proj_matrix)


class GeneralProjectionLearner(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        chain=({'type': 'pca', 'args': {'n_components': 5}},),
        indices=None,
        n_freq=1025,
    ):
        self.chain = chain
        self.indices = indices
        self.n_freq = n_freq

    def fit(self, X, y=None):
        self.scalings_ = learn_chain_proj_matrix(
            X, y, self.chain, indices=self.indices, input_size=self.n_freq
        )
        self.projection = np.dot(self.scalings_, self.scalings_.T)
        return self

    def transform(self, X):
        """Projection within projected space (reduces dimensions)"""
        return np.dot(X, self.scalings_)

    def project(self, X):
        """Projection within original space (remains same dimensions, no reduction)"""
        return np.dot(X, self.projection)


def make_buckets(
    n_buckets=15,
    freqs_weighting=lambda x: x,
    freq_range=DFLT_FREQ_RANGE,
    non_empty_bucket=True,
    reverse=False,
):
    """
    Create greedily buckets starting by aggregating lower frequencies, when the sum of
    the frequencies values so far exceed the number of buckets created times the target
    average value for a single bucket, a new bucket is created with or without the last
    term according to which choice will be closest.

    :param n_buckets: final number of buckets
    :param freqs_weighting: any function assigning a non-negative value to each
        element in the freq_range.
        If given a list, the function will be assumed to be the one to one mapping the
        frequencies to the value in the list

    :param freq_range: the range of frequencies considered, inclusive on both ends
    :param non_empty_bucket: if set to true, all buckets will have at least one element
    :param reverse: if set to true, will start aggregating higher frequencies instead,
        which for an increasing weight
        function will aggregate the higher frequencies into larger bins as typical
    :return: a partition of freq_range

    >>> buckets = make_buckets(
    ... n_buckets=15,
    ... freqs_weighting=lambda x: np.log(x + 0.001),
    ... freq_range=(200, 1000))
    >>> len(buckets) == 15
    True
    >>> buckets[0][0] # the first bucket starts at the first term of freq_range
    200
    >>> buckets[-1][-1] # the last bucket ends at the last term of freq_range - 1
    999

    >>> make_buckets(3, freq_range=(0, 10))
    [[0, 1, 2, 3, 4, 5, 6], [7, 8, 9]]
    >>> make_buckets(3, freqs_weighting=lambda x: x, freq_range=(0, 9))
    [[0, 1, 2, 3, 4, 5], [6, 7, 8]]
    >>> make_buckets(3, freqs_weighting=lambda x: 3, freq_range=(0, 9))
    [[0, 1, 2], [3, 4, 5, 6], [7, 8]]
    >>>
    >>> make_buckets(3, freqs_weighting=lambda x: 1 / (x + 1), freq_range=(0, 9))
    [[0], [1, 2], [3, 4, 5, 6, 7, 8]]
    >>> list(map(len, make_buckets(
    ... 8, freqs_weighting=lambda x: 1 / (x + 1), freq_range=(0, 1000)))
    ... )
    [1, 2, 5, 14, 34, 88, 225, 632]

    """

    # total number of frequencies
    low_freq = freq_range[0]
    high_freq = freq_range[1]
    n_freqs = high_freq - low_freq
    if n_freqs < n_buckets and non_empty_bucket:
        warnings.warn(
            'You asked for more buckets than the number of frequencies available, '
            'some will necessarily be empty'
        )
        non_empty_bucket = False

    # indices of the frequencies
    freq_range = list(range(low_freq, high_freq + 1))
    if reverse:
        freq_range.reverse()
    # get the value of each frequency
    if not hasattr(freqs_weighting, '__iter__'):
        freq_values = list(map(freqs_weighting, freq_range))
    else:
        freq_values = freqs_weighting
    # ideal average sum of bucket value
    target_bucket_sum = sum(freq_values) / n_buckets

    # position of the next term to consider
    position = 0
    # sum of all the existing buckets plus the current bucket under construction
    existing_bucket_sum = 0
    # list of indices for each bucket, empty at moment
    idx_bucket_list = []

    for idx in range(n_buckets):
        # the list of indices in the bucket under construction
        bucket_idx = []
        while existing_bucket_sum < target_bucket_sum * (idx + 1):
            bucket_idx.append(position)
            existing_bucket_sum += freq_values[position]
            # increment the position if we are not running out of terms
            if position < n_freqs - 1:
                position += 1
            # otherwise we stop and return all the buckets, including the one under construction
            # which is non empty by construction
            else:
                idx_bucket_list.append(bucket_idx)
                if low_freq > 0:
                    idx_bucket_list = [
                        [i + low_freq for i in l] for l in idx_bucket_list
                    ]
                return idx_bucket_list

        # if we skipped the loop above, our newly constructed bucket is empty
        # if we specified that empty buckets are not ok, we force one guy in it
        if non_empty_bucket and len(bucket_idx) == 0:
            bucket_idx.append(position)
            idx_bucket_list.append(bucket_idx)
            # increment the position if we are not running out of terms
            if position < n_freqs - 1:
                position += 1
            # otherwise we stop and return all the buckets, including the one under construction
            # which is non empty by construction
            else:
                return idx_bucket_list

        else:
            # we now have a non empty bucket, so we check what is best, including the last added term or no
            total_dif_small = abs(
                existing_bucket_sum
                - freq_values[position]
                - target_bucket_sum * (idx + 1)
            )
            total_diff_large = abs(existing_bucket_sum - target_bucket_sum * (idx + 1))

            # we don't remove the last if it is better not too or if we want non empty buckets
            # and remove it would violate that rule
            if total_diff_large < total_dif_small or (
                len(bucket_idx) < 2 and non_empty_bucket
            ):
                idx_bucket_list.append(bucket_idx)
            # otherwise remove last term
            else:
                idx_bucket_list.append(bucket_idx[:-1])
                # below is to ensure that if the last bucket is empty, we increment the position
                # effectively skipping that position. Otherwise we will end up with a list of empty lists
                if len(bucket_idx[:-1]) > 0:
                    existing_bucket_sum -= freq_values[position]
                    position -= 1
                else:
                    pass

    idx_bucket_list[-1] = freq_range[idx_bucket_list[-1][0] :]
    if low_freq > 0:
        idx_bucket_list = [[i + low_freq for i in l] for l in idx_bucket_list]

    return idx_bucket_list


def _plus_one_inverse(x):
    return 1 / (x + 1)


def frequency_weight_based_bands_matrix(
    n_buckets=15, freqs_weighting=_plus_one_inverse, n_freq=1025, reverse=False
):
    """
    Get a projection matrix of disjoint bands based on a frequency weighting function.

    >>> assert frequency_weight_based_bands_matrix(3, n_freq=9).tolist() == [
    ... [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ... [0.0, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ... [0.0, 0.0, 0.0, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125]
    ... ]

    """
    buckets = make_buckets(
        n_buckets, freqs_weighting, freq_range=(0, n_freq + 1), reverse=reverse
    )
    return make_band_matrix(buckets, n_freq=n_freq)


# class GeneralProjectionLearner(BaseEstimator, TransformerMixin):
#     def __init__(self, chain=({'type': 'pca', 'kwargs': {'n_components': 5}},), indices=None,
#                  input_size=DFLT_INPUT_SIZE):
#         self.chain = chain
#         self.indices = indices
#         self.input_size = input_size
#
#     def fit(self, X, y=None):
#         pre_projection = mk_pre_projection_from_indices(self.indices, self.input_size)
#         self.scalings_ = learn_chain_proj_matrix(X, y, self.chain, pre_projection)
#         return self
#
#     def fit(self, X, y=None):
#         self.scalings_ = learn_chain_proj_matrix(X, y, self.chain, indices=self.indices, n_freq=self.n_freq)
#         self.projection = np.dot(self.scalings_, self.scalings_.T)
#         return self
#
#     def transform(self, X):
#         """Projection within projected space (reduces dimensions)"""
#         return np.dot(X, self.scalings_)
#
#     def project(self, X):
#         """Projection within original space (remains same dimensions, no reduction)"""
#         return np.dot(X, self.projection)
#



"""Snipping: Feature vector quantization"""

from typing import Callable
import itertools
from collections import Counter, defaultdict

import numpy as np
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.cluster import KMeans

from slang.chunkers import fixed_step_chunker
from slang.snip_stats import BayesFactors
from slang.core import Snipper
from slang.util import row_euclidean_distance, mk_callable


class DfltWfToChk:
    def __init__(self, chk_size=2048, chk_step=None):
        if chk_step is None:
            chk_step = chk_size
        self.chk_size = chk_size
        self.chk_step = chk_step

    def fit(self, *args, **kwargs):
        return self  # no fitting required

    def __call__(self, wf):
        yield from fixed_step_chunker(
            wf, chk_size=self.chk_size, chk_step=self.chk_step
        )


@mk_callable('single_transform')
class PcaChkToFv(PCA):
    def __init__(self, n_components=5, **kwargs):
        super().__init__(n_components=n_components, **kwargs)

    # def __call__(self, fv):
    #     return self.transform([fv])[0]


@mk_callable('single_transform')
class LdaChkToFv(LinearDiscriminantAnalysis):
    def __init__(self, n_components=5, **kwargs):
        super().__init__(n_components=n_components, **kwargs)


DfltChkToFv = PcaChkToFv


class FvToSnip:  # TODO: Mixin? ABC?
    # fit: optional
    # __call__: required
    # fvs_to_snips: required?
    # fvs_to_snip_distance_pairs: required?
    @property
    def fv_of_snip(self):
        """array providing representative fv for each snip"""
        raise NotImplemented(
            'Not implemented -- should be implemented in subclass of FvToSnipMixin'
        )
        # return None


from slang.stypes import FV, FVs


# TODO: choose n_clusters from len(X) at fit time
# TODO: post_fit_proc shouldn't be here, but external and optional
@mk_callable('single_predict')
class KMeansFvToSnip(KMeans, FvToSnip):
    stats_of_snip = None  # will be filled when instance is fit

    # cluster_centers_ = None  # will be filled when instance is fit
    # n_clusters = None
    #
    # fv_of_snip = cluster_centers_  # Snipper's api expectation
    # alphabet_size = n_clusters  # Snipper's api expectation

    @property
    def fv_of_snip(self):
        return self.cluster_centers_

    @property
    def alphabet_size(self):
        return self.n_clusters

    def __init__(self, n_clusters=47, **kwargs):
        super().__init__(n_clusters=n_clusters, **kwargs)

    # def __call__(self, fv: FV):
    #     return self.predict([fv])[0]

    def fit(self, fvs: FVs, y=None, sample_weight=None):
        super().fit(fvs, y, sample_weight)
        # self.post_fit_proc(fvs, y, sample_weight=sample_weight)
        return self

    # fvs_to_snips = KMeans.predict  # alias for predict
    # ... replaced (because need to handle iterator of fvs) by:
    def fvs_to_snips(self, fvs, sample_weight=None):
        # return self.predict(list(fvs), sample_weight=sample_weight)
        if isinstance(fvs, (list, tuple, np.ndarray)):
            return self.predict(fvs, sample_weight=sample_weight)
        else:
            return map(lambda x: self.predict([x]), fvs)

    def fvs_to_snip_distance_pairs(self, fvs: FVs):
        """iterator of (snip, distance_to_snip_centroid) pairs"""
        fvs = np.array(list(fvs))
        snips = self.fvs_to_snips(fvs)
        return zip(snips, row_euclidean_distance(fvs, self.fv_of_snip[snips]))

    # def post_fit_proc(self, fvs, y=None, **kwargs):
    #     d = defaultdict(list)
    #     # TODO: Better than a for loop? (d.update(self.snip_and_distance(fvs)) doesn't work, but want something like))
    #     for snip, distance in self.fvs_to_snip_distance_pairs(fvs):
    #         d[snip].append(distance)
    #
    #     # TODO: A numpy array might be more useful here
    #     # self.stats_of_snip = np.empty(self.alphabet_size) * np.nan
    #
    #     self.stats_of_snip = {snip:
    #                               dict(count=len(distances),
    #                                    dist_sum=sum(distances),
    #                                    dist_sum2=sum(np.array(distances) ** 2))
    #                           for snip, distances in d.items()}
    #
    #     return self


class KMeansFvToSnipDist(KMeansFvToSnip):
    def __call__(self, fv: FV):
        return list(self.fvs_to_snip_distance_pairs([fv]))[0]


DfltFvToSnip = KMeansFvToSnip

from collections.abc import Iterable


def is_iterable(x):
    """Similar in nature to :func:`callable`, ``is_iterable`` returns
    ``True`` if an object is `iterable`_, ``False`` if not.
    >>> is_iterable([])
    True
    >>> is_iterable(1)
    False"""
    return isinstance(x, Iterable)


def _assure_pair(iterables):
    """ """
    if hasattr(iterables, '__len__'):
        if len(iterables) == 2:
            return iterables
        elif len(iterables) == 1 and is_iterable(iterables[0]):
            return list(zip(*iterables[0]))
        # TODO: else?... raise? If None on purpose, say it and why!
    else:
        return list(zip(*iterables))


_get_pairs = _assure_pair  # but deprecating


def _assure_zipped(iterables):
    """ """
    if hasattr(iterables, '__len__'):
        if len(iterables) == 1:
            return iterables
        else:
            return zip(*iterables)
    else:
        return zip(*iterables)


def _is_a_tuple_of_aligned_iterables(x):
    return hasattr(x, '__len__') and len(x) == 2


def iterate_over_pairs(pairs):
    if _is_a_tuple_of_aligned_iterables(pairs):
        yield from zip(*pairs)
    else:
        yield from pairs  # assume pairs is an iterable of pairs


class FittableSnipper(Snipper):
    wf_to_chks: Callable
    chk_to_fv: Callable  # TODO: Make a "Callable and Fittable" type
    fv_to_snip: FvToSnip  # TODO: Make a "Callable and Fittable" type

    dflt_cls_of_name: dict = {
        'wf_to_chks': DfltWfToChk,
        'chk_to_fv': PcaChkToFv,
        'fv_to_snip': DfltFvToSnip,
    }

    @classmethod
    def mk_component(cls, obj, role=None, assert_func=None):
        if isinstance(obj, type):
            obj = obj()
        elif obj is None:
            obj = cls.dflt_cls_of_name.get(role)()
        elif not callable(obj):
            assert isinstance(obj, dict)
            obj = cls.dflt_cls_of_name.get(role)(**obj)
        else:
            pass  # as is
        if assert_func is not None:
            assert assert_func(obj)
        return obj

    def __init__(self, wf_to_chks=None, chk_to_fv=None, fv_to_snip=None):

        wf_to_chks = self.mk_component(wf_to_chks, 'wf_to_chks', assert_func=callable)
        chk_to_fv = self.mk_component(chk_to_fv, 'chk_to_fv', assert_func=callable)
        fv_to_snip = self.mk_component(fv_to_snip, 'fv_to_snip', assert_func=callable)

        super().__init__(wf_to_chks, chk_to_fv, fv_to_snip)

        # TODO: Find a more concise way to take care of block above. Like... (but not working)
        # _locals = locals()
        # for obj_role in ['wf_to_chks', 'chk_to_fv', 'fv_to_snip', 'snip_to_score']:
        #     _locals[obj_role] = self.mk_component(_locals[obj_role], obj_role, assert_func=callable)

    # TODO: Make the next three methods more DRY (see also ClassificationSnipper.fit_snip_to_score)
    def fit_wf_to_chks(self, *wfs_tags):
        if hasattr(self.wf_to_chks, 'fit'):
            chks, tags = _assure_pair(
                wfs_tags
            )  # need to generalize to situations with no tags
            self.wf_to_chks.fit(chks, tags)
        return self

    def fit_chk_to_fv(self, *chks_tags):
        if hasattr(self.chk_to_fv, 'fit'):
            chks, tags = _assure_pair(chks_tags)
            self.chk_to_fv.fit(chks, tags)
        return self

    def fit_fv_to_snip(self, *fvs_tags):
        if hasattr(self.fv_to_snip, 'fit'):
            fvs, tags = _assure_pair(fvs_tags)
            self.fv_to_snip.fit(fvs, tags)
        return self

    # TODO: Dagify all this. Too hardcoded. No caching. No flexibility

    def chk_tag_gen(self, wf_tag_gen):
        for wf, tag in wf_tag_gen():
            for chk in self.wf_to_chks(wf):
                yield chk, tag

    def fv_tag_gen(self, wf_tag_gen):
        for chk, tag in self.chk_tag_gen(wf_tag_gen):
            yield self.chk_to_fv(chk), tag

    def snip_tag_gen(self, wf_tag_gen):
        for fv, tag in self.fv_tag_gen(wf_tag_gen):
            yield self.fv_to_snip(fv), tag

    def fit_pipeline(self, wf_tags_gen):
        self.fit_wf_to_chks(wf_tags_gen())
        self.fit_chk_to_fv(self.chk_tag_gen(wf_tags_gen))
        self.fit_fv_to_snip(self.fv_tag_gen(wf_tags_gen))

    def fit(self, X, y=None):
        if y is None:
            wf_tags_gen = lambda: itertools.product(X, [y])
        else:
            wf_tags_gen = lambda: zip(X, y)
        return self.fit_pipeline(wf_tags_gen)


class ClassificationSnipper(FittableSnipper):
    snip_to_score: Callable  # TODO: Make a "Callable and Fittable" type

    dflt_cls_of_name: dict = dict(
        FittableSnipper.dflt_cls_of_name, snip_to_score=BayesFactors
    )

    def __init__(
        self,
        wf_to_chks=DfltWfToChk(),
        chk_to_fv=DfltChkToFv,
        fv_to_snip=DfltFvToSnip,
        snip_to_score=BayesFactors,
    ):
        super().__init__(wf_to_chks, chk_to_fv, fv_to_snip)

        snip_to_score = self.mk_component(
            snip_to_score, 'snip_to_score', assert_func=callable
        )
        self.snip_to_score = snip_to_score

    def fit_snip_to_score(self, *snips_tags):
        if hasattr(self.snip_to_score, 'fit'):
            snips, tags = _assure_pair(snips_tags)
            self.snip_to_score.fit(snips, tags)
        return self

    # TODO: Dagify all this. Too hardcoded. No caching. No flexibility

    def fit_pipeline(self, wf_tags_gen):
        super().fit_pipeline(wf_tags_gen)
        self.fit_snip_to_score(self.snip_tag_gen(wf_tags_gen))


# from slang.snip_stats import ClassifiedMomentsFitter
# from itertools import chain
#
#
# class OutlierSnipper(FittableSnipper):
#     snip_to_score: Callable  # TODO: Make a "Callable and Fittable" type
#
#     dflt_cls_of_name: dict = dict(FittableSnipper.dflt_cls_of_name, snip_to_score=ClassifiedMomentsFitter)
#
#     def __init__(self, wf_to_chks=DfltWfToChk(),
#                  chk_to_fv=DfltChkToFv,
#                  fv_to_snip=DfltFvToSnip,
#                  snip_to_score=ClassifiedMomentsFitter):
#         super().__init__(wf_to_chks, chk_to_fv, fv_to_snip)
#
#         snip_to_score = self.mk_component(snip_to_score, 'snip_to_score', assert_func=callable)
#         self.snip_to_score = snip_to_score
#
#     def fit_snip_to_score(self, *dist_snips):
#         if hasattr(self.snip_to_score, 'fit'):
#             dist, snips = _assure_pair(dist_snips)
#             self.snip_to_score.fit(dist, snips)
#         return self
#
#     # TODO: Dagify all this. Too hardcoded. No caching. No flexibility
#
#     def fit_pipeline(self, wf_gen):
#         # super().fit_pipeline(wf_tags_gen)
#         self.fit_wf_to_chks(wf_gen())
#         self.fit_chk_to_fv(self.chk_tag_gen(wf_tags_gen))
#         fvs = map(self.chk_to_fv, chain.from_iterable(map(self.wf_to_chks,
#         self.fit_fv_to_snip(self.fv_tag_gen(wf_tags_gen))
#
#         self.fit_snip_to_score(self.snip_tag_gen(wf_tags_gen))

# def snips_of_wf(self, wf: Waveform) -> Snips:
#     warn("The name 'snips_of_wf' be replaced by 'wf_to_snips' soon.")
#     for chk in self.wf_to_chks(wf):
#         fv = self.chk_to_fv(chk)
#         yield self.fv_to_snip(fv)
#
# def wf_to_fvs(self, wf: Waveform) -> FVs:
#     for chk in self.wf_to_chks(wf):
#         yield self.chk_to_fv(chk)
#
# def chk_to_snip(self, chk: Chunk) -> Snip:
#     return self.fv_to_snip(self.chk_to_fv(chk))
#
# def wf_to_snips(self, wf: Waveform) -> Snips:
#     for chk in self.wf_to_chks(wf):
#         fv = self.chk_to_fv(chk)
#         yield self.fv_to_snip(fv)


SlangClassifier = ClassificationSnipper  # alias for back compatibility



from slang.stypes import Waveforms, Chunker, Featurizer, Snipper


def just_give_me_any_snipper() -> Snipper:
    pass


def fit_chunker_from_wf_gen(chunker: Chunker, wf_gen: Waveforms) -> Chunker:
    pass


def fit_featuriizer_from_wf_gen(
    featurizer: Featurizer, wf_gen: Waveforms
) -> Featurizer:
    pass



"""Snip statistics functions"""

from collections import defaultdict, Counter
from operator import itemgetter
from dataclasses import dataclass
from functools import reduce
from typing import Iterable, Tuple, Union

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from dol.util import groupby  # TODO: consider changing to use itertools.groupby

from slang.util import lazyprop, running_mean_gen, row_euclidean_distance
from slang.util import snips_to_str as dflt_snips_to_str

GroupId = int
FV = Union[float, Iterable[float]]
GroupedFV = Tuple[GroupId, FV]


# Note: ClassifiedMoments and ClassifiedMomentsFitter untested and paused for now
# @dataclass
# TODO: Review-w-mp
class ClassifiedMoments:
    _zero_var = 33e-33

    def __init__(self, n_classes_: int, initial_count: int = 1):
        self.n_classes_ = n_classes_
        self.initial_count = initial_count
        self.initialize_stats()

    def initialize_stats(self):
        self.count_ = np.ones(self.n_classes_) * self.initial_count
        self.sum_ = np.zeros(self.n_classes_)
        self.sum2_ = np.zeros(self.n_classes_)
        return self

    @property
    def global_count_(self):
        return len(self.count_)

    @property
    def global_sum_(self):
        return sum(self.sum_)

    @property
    def global_sum2_(self):
        return sum(self.sum2_)

    @property
    def global_mean_(self):
        return self.global_sum_ / self.global_count_

    @property
    def global_var_(self):
        return max(
            self._zero_var,
            (self.global_sum2_ / self.global_count_) - self.global_mean_ ** 2,
        )  # TODO: verify

    @property
    def global_std_(self):
        return np.sqrt(self.global_var_)

    @property
    def inverse_frequency_(self):
        return 1 / self.count_

    @property
    def mean_(self):
        return self.sum_ / (self.count_)

    @property
    def var_(self):
        """Variance: E[X**2] - E[X]**2 (https://en.wikipedia.org/wiki/Variance)"""
        return (self.sum2_ / self.count_) - self.mean_ ** 2  #

    def inverse_frequency_for_groups(self, groups):
        return 1 / self.count_[groups]

    def var_for_groups(self, groups):
        count = self.count_[groups]
        return (self.sum2_[groups] / count) - (self.sum_[groups] / count) ** 2

    def predict_within_group(self, gfvs: Iterable[GroupedFV]):
        groups, fvs = list(map(np.array, zip(*gfvs)))
        return np.array(
            list(
                zip(
                    self.inverse_frequency_for_groups(groups),
                    fvs / np.sqrt(self.var_for_groups(groups)),
                )
            )
        )

    def predict_global(self, gfvs: Iterable[GroupedFV]):
        groups, fvs = list(map(np.array, zip(*gfvs)))
        return np.array(
            list(zip(self.inverse_frequency_for_groups(groups), fvs / self.global_std_))
        )

    predict = predict_global

    def __call__(self, gfv: GroupedFV):
        return self.predict([gfv])[0]


# Make ABC or mother class
class ClassifiedMomentsFitter(ClassifiedMoments, BaseEstimator):
    def fit_partial_single(self, gfv: GroupedFV):
        group, fv = gfv
        self.count_[group] += 1
        self.sum_[group] += fv
        self.sum2_[group] += fv ** 2
        return self

    # # TODO: Make this work
    # def __iadd__(self, gfv: GroupedFV):
    #     return self.fit_partial_single(gfv)

    def _group_x_by_y(self, X, y):
        d = defaultdict(list)
        # TODO: Better than a for loop? (d.update(self.snip_and_distance(fvs)) doesn't work, but want something like))
        for label, x in zip(y, X):
            d[label].append(x)
        return d

    def fit_partial(self, gfvs: Iterable[GroupedFV], y=None):
        for gfv in gfvs:
            self.fit_partial_single(gfv)

        return self

    def fit(self, gfvs: Iterable[GroupedFV], y=None):
        if not self.n_classes_:
            self.n_classes_ = np.max(y)
        self.initialize_stats()

        return self.fit_partial(gfvs)


def _is_a_tuple_of_aligned_iterables(x):
    return hasattr(x, '__len__') and len(x) == 2


def iterate_over_pairs(pairs):
    if _is_a_tuple_of_aligned_iterables(pairs):
        yield from zip(*pairs)
    else:
        yield from pairs  # assume pairs is an iterable of pairs


def mk_snips_of_tag_dict(tags, snips):
    return groupby(zip(tags, snips), itemgetter(0), itemgetter(1))


# TODO: Fillna is not additive smoothing, so should make it so?
class TagSnipStats:
    def __init__(self, snips, tags, snips_to_str=None, fillna=0, tag_order=None):
        """

        :param snips: The snips to analyze
        :param tags: The corresponding tags (one per snip)
        :param snips_to_str: A function that outputs a string of characters corresponding to the snip sequence
        :param fillna: The pseudocount that will be used in the probability estimator.
            That is, the value to assign to non-observed (snip, tag) pairs count.
            See https://en.wikipedia.org/wiki/Additive_smoothing
        :param tag_order:
        """
        self.snips = snips
        self.tags = tags
        self.snips_to_str = snips_to_str
        self.fillna = fillna
        self.tag_order = tag_order

    @lazyprop
    def snip_tag_counts(self):
        df = df_of_snip_count_for_tag(
            (self.tags, self.snips), self.snips_to_str, self.fillna, self.tag_order
        )
        if self.tag_order is None:
            self.tag_order = df.columns.values
        total_count = df.sum(axis=1)
        snip_order = (total_count.sort_values(ascending=False)).index.values
        df = df.loc[snip_order]
        return df

    @lazyprop
    def snip_count_for_tag(self):
        return tag_snip_count_dict_from_tags_and_snips(self.tags, self.snips)

    @lazyprop
    def snip_order(self):
        return self.snip_tag_counts.index.values

    @lazyprop
    def log_bayes_factor(self):
        return np.log2(bayes_factor_df_from_snip_count_df(self.snip_tag_counts))

    @lazyprop
    def snips_of_tag(self):
        return mk_snips_of_tag_dict(self.tags, self.snips)

    def scores_for_tag(self, tag, snips=None):
        if snips is None:
            snips = self.snips
        return np.array(list(map(self.log_bayes_factor[tag].loc.__getitem__, snips)))

    def plot_snip_count_for_tag(
        self,
        snips_to_str=None,
        figsize=(14, 10),
        tag_order=None,
        output_fig=False,
        ylabel_rotation=90,
    ):
        return plot_snip_count_for_tag(
            self.snip_count_for_tag,
            snips_to_str=snips_to_str,
            figsize=figsize,
            tag_order=tag_order,
            output_fig=output_fig,
            ylabel_rotation=ylabel_rotation,
        )

    # TODO: Make a decorator that takes care of the `x = x or getattr(self, x, None)` pattern
    def plot_bars_of_tag_snip_stats(
        self,
        figsize=(24, 18),
        output_fig=False,
        ylabel_fontsize=None,
        ylabel_rotation=90,
        tag_order=None,
        snip_order=None,
        snips_to_str=None,
    ):
        tag_order = tag_order or self.tag_order
        snip_order = snip_order or self.snip_order
        snips_to_str = snips_to_str or self.snips_to_str
        bar_plot_of_tag_snip_stats(
            self.log_bayes_factor,
            snips_to_str=snips_to_str,
            figsize=figsize,
            snip_order=snip_order,
            tag_order=tag_order,
            output_fig=output_fig,
            ylabel_fontsize=ylabel_fontsize,
            ylabel_rotation=ylabel_rotation,
        )

    def plot_tag_scores(
        self,
        tags=None,
        snips=None,
        chk_size=1,
        chk_step=1,
        figsize=(24, 18),
        ylabel_fontsize=20,
        ylabel_rotation=0,
        predict_thresh=0,
        normal_style='-k',
        over_predict_style='ob',
    ):
        if tags is None:
            tags = self.tag_order
        if snips is None:
            snips = np.array(reduce(lambda x, y: x + y, self.snips_of_tag.values(), []))

        gen = tag_scores_gen(self.scores_for_tag, tags, snips, chk_size, chk_step)
        tag_snips_cursor = 0
        n_tags = len(tags)
        plt.figure(figsize=figsize)
        for i, (tag, scores) in enumerate(gen, 1):
            scores = np.array(scores)
            x_vals = np.arange(len(scores))

            plt.subplot(n_tags, 1, i)
            plt.plot(scores, normal_style)
            plt.plot([0, len(scores)], [0, predict_thresh], ':', color='k', alpha=0.8)

            over_predict_thresh = scores >= predict_thresh
            plt.plot(
                x_vals[over_predict_thresh],
                scores[over_predict_thresh],
                over_predict_style,
            )

            #     n_tag_snips = len(tss.snips_of_tag[tag])
            #     these_snip_scores = snip_scores[tag_snips_cursor:(tag_snips_cursor + n_tag_snips)]
            #     tag_snips_idx = list(range(tag_snips_cursor, tag_snips_cursor + len(these_snip_scores)))
            #     plt.plot(tag_snips_idx, these_snip_scores, '-o',  color='k');
            #     tag_snips_cursor += n_tag_snips

            plt.axis('tight')
            plt.ylabel(tag, fontsize=ylabel_fontsize, rotation=ylabel_rotation)

    def plot_smoothed_log_bayes_factors(
        self, chk_size=1, chk_step=None, tag_order=None
    ):
        return plot_smoothed_log_bayes_factors(
            self.snips_of_tag,
            self.log_bayes_factor,
            chk_size=chk_size,
            chk_step=chk_step,
            tags=tag_order,
        )

    def tags_and_snips_str_gen(self, tags=None, snips=None, snips_to_str=None):
        snips_to_str = snips_to_str or self.snips_to_str or dflt_snips_to_str
        if tags is None and snips is None:
            snips_of_tag = self.snips_of_tag
        else:
            snips_of_tag = mk_snips_of_tag_dict(tags, snips)

        for tag, snips in snips_of_tag.items():
            yield tag, snips_to_str(snips)

    def mk_tags_and_snips_str_string(
        self,
        tags=None,
        snips=None,
        snips_to_str=None,
        tag_snips_format_str='{:<21}: {}\n\n',
    ):
        s = ''
        for tag, snips_str in self.tags_and_snips_str_gen(tags, snips, snips_to_str):
            s += tag_snips_format_str.format(tag, snips_str)
        return s

    def print_tags_and_snips_str(
        self,
        tags=None,
        snips=None,
        snips_to_str=None,
        tag_snips_format_str='{:<21}: {}\n\n',
    ):
        print(
            self.mk_tags_and_snips_str_string(
                tags, snips, snips_to_str, tag_snips_format_str
            )
        )


# class SnipStats:
#     """Compute frequency and centroid distance statistics on snips
#     """
#
#     def __init__(self, pseudocount=0, tag_order=None, alphabet_size=None):
#         self.pseudocount = pseudocount
#         self.tag_order = tag_order
#         self._alphabet_size = alphabet_size  # TODO: Use to tell TagSnipStats it should fill until there
#
#     @lazyprop
#     def alphabet_size(self):
#         return self._alphabet_size or self.log_bayes_factor_.index.max()
#
#     def fit(self, snips, tags=None):
#         self.tag_snip_stats = TagSnipStats(snips, tags, fillna=self.pseudocount, tag_order=self.tag_order)
#         self.log_bayes_factor_ = self.tag_snip_stats.log_bayes_factor.sort_index()
#         self.classes_ = self.tag_snip_stats.tag_order
#         return self


class BayesFactors:
    """BayesFactors classifier with sklearn-like interface. predict_probas are the log2 of the bayes factor.
    It is sklearn-like, but to avoid depending on sklearn, it isn't a subclass of BaseEstimator and ClassifierMixin.
    It just has the essentials of the classifier: a `fit`, a `predict_proba` and a derived `predict` method, and
    a `classes_` attribute that indices the columns of the `predict_proba` matrix.
    """

    def __init__(self, pseudocount=0, tag_order=None, alphabet_size=None):
        self.pseudocount = pseudocount
        self.tag_order = tag_order
        self._alphabet_size = (
            alphabet_size  # TODO: Use to tell TagSnipStats it should fill until there
        )

    @lazyprop
    def alphabet_size(self):
        return self._alphabet_size or self.log_bayes_factor_.index.max()

    def fit(self, snips, tags):
        self.tag_snip_stats = TagSnipStats(
            snips, tags, fillna=self.pseudocount, tag_order=self.tag_order
        )
        self.log_bayes_factor_ = self.tag_snip_stats.log_bayes_factor.sort_index()
        self.classes_ = self.tag_snip_stats.tag_order
        return self

    @classmethod
    def from_params(cls, log_bayes_factor_, classes_=None):
        self = cls()
        if classes_ is None:
            classes_ = log_bayes_factor_.columns.values
        self.log_bayes_factor_ = log_bayes_factor_[
            classes_
        ]  # To assert columns contents and ensure order
        self.classes_ = classes_

    def scores_for_tag(self, tag, snips):
        return np.array(list(map(self.log_bayes_factor_[tag].loc.__getitem__, snips)))

    def predict_proba(self, snips):
        scores_for_tag = self.scores_for_tag
        return np.vstack(list(scores_for_tag(tag, snips) for tag in self.classes_)).T

    def predict(self, snips):
        """
        Predict class labels for each snip of snips sequence.
        """
        indices = self.predict_proba(snips).argmax(axis=1)
        return self.classes_[indices]

    def _assert_sanity(self):
        assert set(np.diff(sorted(self.log_bayes_factor_.index))) == {
            1
        }, 'some snips are missing!'

    _diagnosis = _assert_sanity  # but deprecating _diagnosis

    # TODO: self.snip_to_score([0]) blows up, but self.snip_to_score([0, 1, 2]) gives me a score (should blow up)
    def __call__(self, snip):
        return self.predict_proba([snip])[0]


def mk_model_caller(kind='asis'):
    if kind == 'predict_proba':

        def model_caller(self, snip):
            return self.predict_proba([snip])[0]

    elif kind == 'tag_probs':

        def model_caller(self, snip):
            return {
                tag: prob
                for tag, prob in zip(self.classes_, self.predict_proba([snip])[0])
            }

    elif kind == 'predict':

        def model_caller(self, snip):
            return self.predict([snip])[0]

    else:
        raise ValueError(f'Unknown kind: {kind}')
    return model_caller


class PredictProbaBF(BayesFactors):
    def __call__(self, snip):
        return self.predict_proba([snip])[0]


class TagProbsBF(BayesFactors):
    def __call__(self, snip):
        return {
            tag: prob for tag, prob in zip(self.classes_, self.predict_proba([snip])[0])
        }


class PredictBF(BayesFactors):
    def __call__(self, snip):
        return self.predict([snip])[0]


def tag_slice_iter_from_slices_of_tag_dict(slices_of_tag):
    """
    Get an iterator of (tag, (bt, tt)) pairs
    :param slices_of_tag: a {tag: [(bt, tt),...], ...} dict listing slices annotated by tags
    :return: a tag, (bt, tt) iterator
    """
    for tag, slices in slices_of_tag.items():
        for sl in slices:
            yield tag, sl


# def snip_count_for_tag_from_wf_and_tag_slice_iter(wf, tag_slice_iter, snips_of_wf, sr=None):
#     """
#
#     :param wf: waveform of oto.sound.audio.Sound object
#     :param tag_slice_iter:
#     :param snips_of_wf:
#     :param sr:
#     :return:
#     """
#
#     # from slang.utils.audio_core import Sound
#     snip_count_for_tag = defaultdict(Counter)
#
#     for tag, sl in tag_slice_iter:
#         chk = sound[slice(*sl)].wf  # wf chunk for sl slice
#         chk_snips = snips_of_wf(chk)
#         snip_count_for_tag[tag].update(chk_snips)
#
#     return dict(snip_count_for_tag)


def tag_snip_count_dict_from_tags_and_snips(tags, snips):
    snip_count_for_tag = defaultdict(dict)

    for (tag, snip), count in Counter(zip(tags, snips)).items():
        snip_count_for_tag[tag][snip] = count

    return dict(snip_count_for_tag)


def df_of_snip_count_for_tag(
    snip_count_for_tag, snips_to_str=None, fillna=0, tag_order=None
):
    """
    A df representation of snip_count_for_tag
    :param snip_count_for_tag: {tag: {snip: count, ...},...} dict
    :param snips_to_str: A function that transforms snip lists into strings (mapping each snip to a character)
    :param fillna: What to fill missing values with
    :param tag_order: Serves both to specify an order of the tags, and to specify a subset of tags if we don't want all
    :return: A dataframe of snip (in rows) counts for each tag (in columns)
    """
    if isinstance(snip_count_for_tag, tuple) and len(snip_count_for_tag) == 2:
        snip_count_for_tag = tag_snip_count_dict_from_tags_and_snips(
            *snip_count_for_tag
        )

    df = pd.DataFrame(snip_count_for_tag).fillna(fillna)
    if tag_order is not None:
        df = df[tag_order]
    df.index.names = ['snip']
    if snips_to_str is not None:
        df = df.reset_index(drop=False)
        df['snip'] = list(snips_to_str(df.index.values))
        df = df.set_index('snip')
    return df


def snip_order_from_snip_count_df(snip_count_df):
    total_count = snip_count_df.sum(axis=1)
    return (total_count.sort_values(ascending=False)).index.values


def bayes_factor_df_from_snip_count_df(snip_count_df):
    smoothed_count = snip_count_df + 1
    total_snip_count = smoothed_count.sum(axis=1)
    prob_given_tag = smoothed_count / smoothed_count.sum(axis=0)
    # prob_given_not_tag = total_snip_count.sub(smoothed_count, level=0, fill_value=0, axis=0)
    prob_given_not_tag = -smoothed_count.sub(total_snip_count, axis=0)
    prob_given_not_tag /= prob_given_not_tag.sum(axis=0)
    return prob_given_tag / prob_given_not_tag


def log_bayes_factor_bayes_factor_df_from_snip_count_df(snip_count_df):
    return np.log2(bayes_factor_df_from_snip_count_df(snip_count_df))


def bar_plot_of_tag_snip_stats(
    snip_stats_for_tag,
    snips_to_str=None,
    figsize=(14, 10),
    snip_order=None,
    tag_order=None,
    output_fig=False,
    ylabel_fontsize=None,
    ylabel_rotation=90,
):
    """
    Multiplot of snip count bars for each tag (in a different row). First row is the total count for each snip.
    :param snip_count_for_tag: {tag: {snip: count, ...},...} nested dict
    :param snips_to_str: A function that transforms snip lists into strings (mapping each snip to a character)
    :param figsize:
    :param output_fig:
    :param tag_order: Serves both to specify an order of the tags, and to specify a subset of tags if we don't want all
    :param ylabel_rotation: Will be applied to the ylabel
    :return:
    """
    if figsize is not None:
        fig = plt.figure(figsize=figsize)
    else:
        fig = plt.gcf()

    if snips_to_str is None:
        snips_to_str = lambda snips: list(map(str, snips))

    if not isinstance(snip_stats_for_tag, pd.DataFrame):
        snip_stats_for_tag = pd.DataFrame(snip_stats_for_tag)

    if snip_order is not None:
        snip_stats_for_tag = snip_stats_for_tag.loc[snip_order, :]

    if tag_order is not None:
        snip_stats_for_tag = snip_stats_for_tag[tag_order]

    n_tags = len(snip_stats_for_tag.columns)
    n_snips = len(snip_stats_for_tag)

    ax_list = list()
    for i, tag in enumerate(snip_stats_for_tag.columns, 1):
        sr = snip_stats_for_tag[tag]
        ax = plt.subplot(n_tags + 1, 1, i)
        ax_list.append(ax)
        positive_lidx = sr > 0
        sr.plot(kind='bar', color=positive_lidx.map({True: 'b', False: '#D3D3D3'}))

        h = plt.ylabel(tag)
        h.set_rotation(ylabel_rotation)
        if ylabel_fontsize is not None:
            h.set_fontsize(ylabel_fontsize)

        plt.xlabel('')
        if i == 1:
            plt.xticks(
                list(range(n_snips)), snips_to_str(snip_stats_for_tag.index.values)
            )
            ax.xaxis.tick_top()
        else:
            plt.xticks(list(range(n_snips)), ' ' * n_snips)

    plt.xticks(list(range(n_snips)), snips_to_str(snip_stats_for_tag.index.values))

    if n_snips < 50:
        for ax in ax_list:
            # ax.grid(True, axis='x')
            ax.grid(True)

    if output_fig:
        return fig


def plot_snip_count_for_tag(
    snip_count_for_tag,
    snips_to_str=None,
    figsize=(14, 10),
    tag_order=None,
    output_fig=False,
    ylabel_fontsize=None,
    ylabel_rotation=90,
):
    """
    Multiplot of snip count bars for each tag (in a different row). First row is the total count for each snip.
    :param snip_count_for_tag: {tag: {snip: count, ...},...} nested dict
    :param snips_to_str: A function that transforms snip lists into strings (mapping each snip to a character)
    :param figsize:
    :param output_fig:
    :param tag_order: Serves both to specify an order of the tags, and to specify a subset of tags if we don't want all
    :param ylabel_rotation: Will be applied to the ylabel
    :return:
    """
    if figsize is not None:
        fig = plt.figure(figsize=figsize)
    else:
        fig = plt.gcf()

    if snips_to_str is None:
        snips_to_str = lambda snips: list(map(str, snips))

    df = df_of_snip_count_for_tag(snip_count_for_tag, fillna=0, tag_order=tag_order)
    n_tags = len(df.columns)
    df['ALL'] = df.sum(axis=1)
    df = df.sort_values('ALL', ascending=False)
    n_snips = len(df)

    ax_list = list()
    for i, tag in enumerate(df.columns, 1):
        sr = df[tag]
        ax = plt.subplot(n_tags + 1, 1, i)
        ax_list.append(ax)
        sr.plot(kind='bar')
        h = plt.ylabel(tag)
        h.set_rotation(ylabel_rotation)
        if ylabel_fontsize is not None:
            h.set_fontsize(ylabel_fontsize)
        plt.xlabel('')
        if i == 1:
            plt.xticks(list(range(n_snips)), snips_to_str(df.index.values))
            ax.xaxis.tick_top()
        else:
            plt.xticks(list(range(n_snips)), ' ' * n_snips)

    plt.xticks(list(range(n_snips)), snips_to_str(df.index.values))

    for ax in ax_list:
        ax.grid(True, axis='x')

    if output_fig:
        return fig


def plot_tag_scores_for_snips(
    snips_of_tag,
    snip_tag_score_df,
    tag_order=None,
    smoothing_window_size=1,
    figsize=(24, 18),
    ylabel_fontsize=15,
    ylabel_rotation=0,
):
    assert isinstance(
        snip_tag_score_df, pd.DataFrame
    ), 'isinstance(snip_tag_score_df, pd.DataFrame)'

    def scores_of_snips(snips, tag):
        return list(map(snip_tag_score_df[tag].loc.__getitem__, snips))

    if tag_order is None:
        tag_order = list(snips_of_tag.keys())
    n_tags = len(tag_order)

    all_snips = reduce(lambda x, y: x + y, snips_of_tag.values(), [])

    plt.figure(figsize=figsize)

    tag_snips_cursor = 0
    for i, tag in enumerate(tag_order, 1):
        plt.subplot(n_tags, 1, i)

        snip_scores = list(
            running_mean_gen(scores_of_snips(all_snips, tag), smoothing_window_size)
        )
        plt.plot(snip_scores, '-')
        plt.plot([0, len(snip_scores)], [0, 0], ':k')

        n_tag_snips = len(snips_of_tag[tag])
        these_snip_scores = snip_scores[
            tag_snips_cursor : (tag_snips_cursor + n_tag_snips)
        ]
        tag_snips_idx = list(
            range(tag_snips_cursor, tag_snips_cursor + len(these_snip_scores))
        )
        plt.plot(tag_snips_idx, these_snip_scores, 'k-')
        tag_snips_cursor += n_tag_snips

        plt.axis('tight')
        plt.ylabel(tag, fontsize=ylabel_fontsize, rotation=ylabel_rotation)


def tags_and_snips_to_snip_of_tag(tags, snips):
    snips_of_tag = defaultdict(list)

    for tag, snip in zip(tags, snips):
        snips_of_tag[tag].append(snip)

    return dict(snips_of_tag)


# def _tag_order_from_df(df_with_tags_as_columns_and_snips_as_indices):
#     snip_order = df_with_tags_as_columns_and_snips_as_indices.index.values
#     tag_order = df_with_tags_as_columns_and_snips_as_indices.columns.values
#     return snip_order, tag_order


# def scores_of_snips(tag, snips, snip_log_bayes_factor_of_tag):
#     lbf_for_tag = snip_log_bayes_factor_of_tag[tag]
#     if isinstance(snip_log_bayes_factor_of_tag, pd.DataFrame):
#         log_bayes_factor_for_snip = lbf_for_tag.loc.__getitem__
#     elif hasattr(lbf_for_tag, '__getitem__'):
#         log_bayes_factor_for_snip = lbf_for_tag.__getitem__
#     else:
#         assert callable(lbf_for_tag), "At this point lbf_for_tag can only be callable"
#         log_bayes_factor_for_snip = lbf_for_tag
#
#     return list(map(log_bayes_factor_for_snip, snips))


def tag_scores_gen(scores_for_tag, tags, snips, chk_size=1, chk_step=1):
    if isinstance(tags, (str, int, float)):
        tags = [tags]
    for tag in tags:
        yield tag, list(
            running_mean_gen(scores_for_tag(tag, snips), chk_size, chk_step)
        )


def plot_tag_scores(
    scores_for_tag,
    tags,
    snips,
    chk_size=1,
    chk_step=1,
    figsize=(24, 18),
    ylabel_fontsize=20,
    ylabel_rotation=0,
    predict_thresh=0,
    normal_style='-k',
    over_predict_style='ob',
):
    gen = tag_scores_gen(scores_for_tag, tags, snips, chk_size, chk_step)
    tag_snips_cursor = 0
    n_tags = len(tags)
    plt.figure(figsize=figsize)
    for i, (tag, scores) in enumerate(gen, 1):
        scores = np.array(scores)
        x_vals = np.arange(len(scores))

        plt.subplot(n_tags, 1, i)
        plt.plot(scores, normal_style)
        plt.plot([0, len(scores)], [0, predict_thresh], ':', color='k', alpha=0.8)

        over_predict_thresh = scores >= predict_thresh
        plt.plot(
            x_vals[over_predict_thresh], scores[over_predict_thresh], over_predict_style
        )

        #     n_tag_snips = len(tss.snips_of_tag[tag])
        #     these_snip_scores = snip_scores[tag_snips_cursor:(tag_snips_cursor + n_tag_snips)]
        #     tag_snips_idx = list(range(tag_snips_cursor, tag_snips_cursor + len(these_snip_scores)))
        #     plt.plot(tag_snips_idx, these_snip_scores, '-o',  color='k');
        #     tag_snips_cursor += n_tag_snips

        plt.axis('tight')
        plt.ylabel(tag, fontsize=ylabel_fontsize, rotation=ylabel_rotation)


def scores_of_snips(tag, snips, snip_log_bayes_factor_of_tag):
    return list(map(snip_log_bayes_factor_of_tag[tag].loc.__getitem__, snips))


def plot_smoothed_log_bayes_factors(
    snips_of_tag,
    snip_log_bayes_factor_of_tag,
    chk_size=1,
    chk_step=1,
    tags=None,
    ylabel_fontsize=15,
    ylabel_rotation=0,
):
    # dflt_snip_order, dflt_tag_order = _tag_order_from_df(snip_tag_counts)
    if tags:
        snip_log_bayes_factor_of_tag = snip_log_bayes_factor_of_tag[tags]
    else:
        tags = list(snip_log_bayes_factor_of_tag.columns)
    n_tags = len(tags)

    all_snips = reduce(lambda x, y: x + y, snips_of_tag.values(), [])

    plt.figure(figsize=(24, 18))

    tag_snips_cursor = 0
    for i, tag in enumerate(tags, 1):
        plt.subplot(n_tags, 1, i)
        snip_scores = list(
            running_mean_gen(
                scores_of_snips(tag, all_snips, snip_log_bayes_factor_of_tag),
                chk_size,
                chk_step,
            )
        )
        plt.plot(snip_scores, '-')
        plt.plot([0, len(snip_scores)], [0, 0], ':k')

        n_tag_snips = len(snips_of_tag[tag])
        these_snip_scores = snip_scores[
            tag_snips_cursor : (tag_snips_cursor + n_tag_snips)
        ]
        tag_snips_idx = list(
            range(tag_snips_cursor, tag_snips_cursor + len(these_snip_scores))
        )
        plt.plot(tag_snips_idx, these_snip_scores, 'k-')
        tag_snips_cursor += n_tag_snips

        plt.axis('tight')
        plt.ylabel(tag, fontsize=ylabel_fontsize, rotation=ylabel_rotation)


def snip_scores_from_lookup(snips, snip_to_score):
    if isinstance(snip_to_score, (pd.Series, dict)):
        snip_to_score = snip_to_score.__getitem__
    elif isinstance(snip_to_score, pd.DataFrame):
        _snip_to_score = {
            k: snip_to_score[k].loc.__getitem__ for k in list(snip_to_score.columns)
        }
        snip_to_score = lambda snip: {k: lookup(snip) for k, lookup in _snip_to_score}
    return map(snip_to_score, snips)

    # assert isinstance(snip_to_score, pd.DataFrame), \
    #     "snip_to_score needs to be DataFrame whose index values are snips and columns are the different score kinds " \
    #     "you want to compute"



"""Chunker functions"""

from itertools import islice, chain
from typing import Callable
from functools import partial

inf = float('inf')

DFLT_CHK_SIZE = 2048

try:
    import numpy as np

    NUMPY_PRESENT = True
except (ImportError, ModuleNotFoundError):
    NUMPY_PRESENT = False


def fannout_chunker(
    objects,
    chunker,
    *,
    extract_chunkable: Callable,
    extract_info: Callable,
    include_chunk_key: bool = False
):
    """Chunk parts of an iterable of objects.

    Transform things like iterables of ``(info1, iterable, info2)`` to iterables
    ``(info2, iterable_chk_0), (info2, iterable_chk_1), ...``.

    :param objects: The iterable of objects that contain both the iterable to chunk and
    the extra info
    :param chunker: The chunker. A function taking the iterable to chunk and returns
    an iterable of chunks
    :param extract_chunkable: The function that extracts the iterable to be chunked
    :param extract_info: The function that extracts from the object the extra info to
    yield for each chunk
    :return: A generator of ``(info, chunk),...`` pairs or ``(chk_idx, (info, chunk))``
    pairs.

    Most of the time you'll be dealing with tuple or dicts as objects, so
    ``operator.itemgetter`` is your friend for creating the extractors.

    >>> from functools import partial
    >>> from operator import itemgetter
    >>> tuple_chunker = partial(
    ...     fannout_chunker,
    ...     chunker=partial(fixed_step_chunker, chk_size=2),
    ...     extract_chunkable=itemgetter(0),
    ...     extract_info=itemgetter(1))
    >>> objects = [([1,2,3,4], 'alice'), ([5,6,7], 'bob')]
    >>> list(tuple_chunker(objects))
    [('alice', [1, 2]), ('alice', [3, 4]), ('bob', [5, 6])]

    If you wanted to have get an index for your chunk as well, you can specify
    ``include_chunk_key=True``:

    >>> tuple_chunker_w_chk_index = partial(tuple_chunker, include_chunk_key=True)
    >>> list(tuple_chunker_w_chk_index(objects))
    [(0, ('alice', [1, 2])), (1, ('alice', [3, 4])), (0, ('bob', [5, 6]))]

    >>> dict_chunker = partial(
    ...     tuple_chunker,
    ...     extract_chunkable=itemgetter('wf'),
    ...     extract_info=itemgetter('tag'),
    ... )
    >>> objects = [
    ...     {'wf': [1,2,3,4], 'tag': 'alice', 'who_cares': 'about_this'},
    ...     {'wf': [5,6,7], 'tag': 'bob', 'or': 'this'}
    ... ]
    >>> list(dict_chunker(objects))
    [('alice', [1, 2]), ('alice', [3, 4]), ('bob', [5, 6])]

    The output format is fixed to ``(info, chunk),...`` or ``(chk_idx, (info, chunk))``.
    To get a different output, write egress function and apply it to output
    systematically. For example:

    >>> from i2 import Pipe
    >>> egress = lambda tup: {'tag': tup[0], 'chk': tup[1]}
    >>> new_dict_chunker = Pipe(dict_chunker, lambda x: map(egress, x))
    >>> assert list(new_dict_chunker(objects)) == [
    ...     {'tag': 'alice', 'chk': [1, 2]},
    ...     {'tag': 'alice', 'chk': [3, 4]},
    ...     {'tag': 'bob', 'chk': [5, 6]}
    ... ]
    """
    for obj in objects:
        iterable_to_chunk = extract_chunkable(obj)
        info = extract_info(obj)
        for chk_idx, chk in enumerate(chunker(iterable_to_chunk)):
            chk_and_info = (info, chk)
            if include_chunk_key:
                yield chk_idx, chk_and_info
            else:
                yield chk_and_info


def _matrix_of_chunks(a, chk_size):
    n = len(a)
    return np.reshape(a[: (n - (n % chk_size))], (-1, chk_size))


def _chunk_with_zip(a, chk_size):
    return zip(*([iter(a)] * chk_size))


def mk_chunker(chk_size=DFLT_CHK_SIZE, chk_step=None, *, use_numpy_reshape=None):
    """
    A fixed step chunker factory. Makes a function that takes an iterable and produces
    fixed-size fixed-step chunks of it's elements.

    This function-making function will try to choose an optimized chunker for you
    depending on the parameters and environment (if you have numpy and
    chk_size==chk_step, it's the fastest).

    Will also add attributes that exhibit the parameters of the chunker.

    Note though, that as a tradeoff, you may get numpy arrays, tuples, or lists as the
    type that is yield.

    :param chk_size: Size of chunk (default 2048)
    :param chk_step: Size of step (step of sliding window). If not specified, will be
    taken to be chk_size
    :param use_numpy_reshape: If None (default), will use numpy (reshape) if numpy is
    importable.
        If True, will try to use numpy.reshape systematically.
        If False, will not use numpy.reshape, even if numpy present.
    :return: A generator of chunks (numpy.arrays, tuples, or lists, depending on the context)

    >>> a = range(6)
    >>> chunker = mk_chunker(3)
    >>> list(chunker(a))
    [array([0, 1, 2]), array([3, 4, 5])]
    >>> list(mk_chunker(2)(a))
    [array([0, 1]), array([2, 3]), array([4, 5])]
    >>> list(tuple(x) for x in mk_chunker(1)(a))
    [(0,), (1,), (2,), (3,), (4,), (5,)]
    >>>
    >>> chunker = mk_chunker(4, 2)
    >>> list(chunker(a))
    [[0, 1, 2, 3], [2, 3, 4, 5]]

    See that ``chunker`` has parameters describing it:

    >>> chunker.chk_size
    4
    >>> chunker.chk_step
    2
    """

    chk_step = chk_step or chk_size  # default to chk_size == chk_step

    if chk_step == chk_size:
        if (use_numpy_reshape is None) or (use_numpy_reshape is True):
            if NUMPY_PRESENT:
                use_numpy_reshape = True
            else:
                if use_numpy_reshape is True:
                    raise ModuleNotFoundError(
                        "You don't have numpy installed, so you can't a reshape chunker"
                    )
                use_numpy_reshape = False

        if use_numpy_reshape:
            chunker = partial(_matrix_of_chunks, chk_size=chk_size)
        else:
            chunker = partial(_chunk_with_zip, chk_size=chk_size)

    else:
        chunker = partial(fixed_step_chunker, chk_size=chk_size, chk_step=chk_step)

    # Adding the chk_size and chk_step attributes to the chunker

    chunker.chk_size = chk_size
    chunker.chk_step = chk_step

    return chunker


DFLT_CHUNKER = mk_chunker(chk_size=DFLT_CHK_SIZE)


def simple_fixed_step_chunker(it, chk_size, chk_step=None):
    """
    Generates chunks of fixed size and step.
    Yields chunks as lists.
    """
    from itertools import islice

    if chk_step is None:
        chk_step = chk_size

    if chk_step == chk_size:
        yield from map(list, zip(*([iter(it)] * chk_step)))
    elif chk_step < chk_size:

        chk = list(islice(it, chk_size))

        while len(chk) == chk_size:
            yield chk
            chk = chk[chk_step:] + list(islice(it, chk_step))

    else:
        chk = list(islice(it, chk_size))
        gap = chk_step - chk_size

        while len(chk) == chk_size:
            yield chk
            chk = list(islice(it, gap, gap + chk_size))


def _validate_chk_size(chk_size):
    assert (
        isinstance(chk_size, int) and chk_size > 0
    ), 'chk_size should be a positive interger'


def _validate_chk_size_and_step(chk_size, chk_step):
    _validate_chk_size(chk_size)
    if chk_step is None:
        chk_step = chk_size
    assert (
        isinstance(chk_step, int) and chk_step > 0
    ), 'chk_step should be a positive integer'
    return chk_size, chk_step


def _validate_fixed_step_chunker_args(chk_size, chk_step, start_at, stop_at):
    chk_size, chk_step = _validate_chk_size_and_step(chk_size, chk_step)

    if start_at is None:
        start_at = 0
    if stop_at is not None:
        assert stop_at > start_at, 'stop_at should be larger than start_at'
        if stop_at is not inf:
            assert isinstance(stop_at, int), 'stop_at should be an integer'

    # checking a few things
    assert isinstance(start_at, int), 'start_at should be an integer'
    assert start_at >= 0, 'start_at should be a non negative integer'
    return chk_step, start_at


def rechunker(
    chks, chk_size, chk_step=None, start_at=None, stop_at=None, return_tail=False
):
    """Takes an iterable of chks and produces another iterable of chunks.
    The chunks generated by the input chks iterable is assumed to be gap-less and without overlap,
    but these do not need to be of fixed size.
    The output will be though.
    """
    yield from fixed_step_chunker(
        chain.from_iterable(chks), chk_size, chk_step, start_at, stop_at, return_tail
    )


def fixed_step_chunker(
    it, chk_size, chk_step=None, start_at=None, stop_at=None, return_tail=False
):
    """
      a function to get (an iterator of) segments (bt, tt) of chunks from an iterator (or list)
      of the for [it_1, it_2...], given a chk_size, chk_step, and a start_at and a stop_at.
      The start_at, stop_at act like slices indices for a list: start_at is included and stop_at is excluded

      :param it: iterator of elements of any type
      :param chk_size: length of the chunks
      :param chk_step: step between chunks
      :param start_at: index of the first term of the iterator at which we begin building the chunks (inclusive)
      :param stop_at: index of the last term from the iterator included in the chunks
      :param return_tail: if set to false, only the chunks with max element with index less than stop_at are yielded
      if set to true, any chunks with minimum index value no more than stop_at are returned but they contain
      term with index no more than stop_at
      :return: an iterator of the chunks

      1) If stop_at is not None and return_tail is False:
         will return all full chunks with maximum element index less than stop_at
         or until the iterator is exhausted. Only full chunks are returned here.

      2) If stop_at is not None and return_tail is True:
         will return all full chunks as above along with possibly cut off chunks
         containing one term whose index is stop_at-1 or one (last) term which is the
         last element of it

      3) If stop_at is None and return_tail is False:
         will return all full chunks with maximum element index less or equal to the last
         element of it

      4) If stop_at is None and return_tail is True:
         will return all full chunks with maximum element index less or equal to the last
         element of it plus cut off chunks whose maximum term index is the last term of it

        # testing chk_step < chk_size with return_tail=TRUE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=1, start_at=2, stop_at=5, return_tail=True)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[3, 4, 5], [4, 5], [5]]

        # testing chk_step < chk_size with return_tail=FALSE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=1, start_at=2, stop_at=5, return_tail=False)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[3, 4, 5]]

        # testing chk_step < chk_size with return_tail=TRUE, stop and start_at PRESENT
        # and stop_at LARGER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=1, start_at=1, stop_at=20, return_tail=True)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9], [8, 9, 10], [9, 10, 11], [10, 11, 12], [11, 12, 13], [12, 13, 14], [13, 14, 15], [14, 15, 16], [15, 16], [16]]

        # testing chk_step < chk_size with return_tail=FALSE, stop and start_at PRESENT
        # and stop_at LARGER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=1, start_at=1, stop_at=20, return_tail=False)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9], [8, 9, 10], [9, 10, 11], [10, 11, 12], [11, 12, 13], [12, 13, 14], [13, 14, 15], [14, 15, 16]]

        # testing chk_step = chk_size with return_tail=TRUE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=3, start_at=1, stop_at=7, return_tail=True)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4], [5, 6, 7]]

        # testing chk_size > len(it) with return_tail=False, no stop_at or start_at
        >>> f = lambda it: fixed_step_chunker(it, chk_size=30, chk_step=3, start_at=None, stop_at=None, return_tail=False)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        []

        # testing chk_size > len(it) with return_tail=True, no stop_at or start_at
        >>> f = lambda it: fixed_step_chunker(it, chk_size=30, chk_step=3, start_at=None, stop_at=None, return_tail=True)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16], [4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16], [7, 8, 9, 10, 11, 12, 13, 14, 15, 16], [10, 11, 12, 13, 14, 15, 16], [13, 14, 15, 16], [16]]

        # testing chk_step > chk_size with return_tail=TRUE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=1, stop_at=7, return_tail=True)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4], [6, 7]]

        # testing chk_step > chk_size with return_tail=FALSE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=1, stop_at=7, return_tail=False)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4]]

        # testing chk_step > chk_size with return_tail=FALSE, stop and start_at NOT PRESENT
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=None, stop_at=None, return_tail=False)
        >>> it = range(1, 17, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[1, 2, 3], [5, 6, 7], [9, 10, 11], [13, 14, 15]]

        # testing chk_step > chk_size with return_tail=TRUE, stop and start_at NOT PRESENT
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=None, stop_at=None, return_tail=True)
        >>> it = range(1, 19, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[1, 2, 3], [5, 6, 7], [9, 10, 11], [13, 14, 15], [17, 18]]

        # testing chk_step > chk_size with return_tail=TRUE, stop and start_at NOT PRESENT
        # with negative values in the iterator
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=None, stop_at=None, return_tail=True)
        >>> it = range(-10, 19, 1)
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[-10, -9, -8], [-6, -5, -4], [-2, -1, 0], [2, 3, 4], [6, 7, 8], [10, 11, 12], [14, 15, 16], [18]]

        # testing chk_step > chk_size with return_tail=TRUE, stop and start_at NOT PRESENT
        # with items of various types in the iterator
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=2, start_at=None, stop_at=None, return_tail=True)
        >>> it = ['a', 3, -10, 9.2, str, [1,2,3], set([10,20])]
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [['a', 3, -10], [-10, 9.2, <class 'str'>], [<class 'str'>, [1, 2, 3], {10, 20}], [{10, 20}]]
       """

    chk_step, start_at = _validate_fixed_step_chunker_args(
        chk_size, chk_step, start_at, stop_at
    )

    if chk_step == chk_size and not return_tail:
        yield from map(list, zip(*([iter(it)] * chk_step)))
    elif chk_step < chk_size:

        it = islice(it, start_at, stop_at)
        chk = list(islice(it, chk_size))

        while len(chk) == chk_size:
            yield chk
            chk = chk[chk_step:] + list(islice(it, chk_step))

    else:
        it = islice(it, start_at, stop_at)
        chk = list(islice(it, chk_size))
        gap = chk_step - chk_size

        while len(chk) == chk_size:
            yield chk
            chk = list(islice(it, gap, gap + chk_size))

    if return_tail:
        while len(chk) > 0:
            yield chk
            chk = chk[chk_step:]


#
# from collections import deque
# from itertools import islice, chain
#
#
# class nothing: ...
#
#
# strip_nothing_out = lambda item: tuple(filter(lambda x: x is not nothing, item))
#
#
# def new_fixed_step_chunker(it, chk_size, chk_step=None, start_at=None, stop_at=None, return_tail=False):
#     chk_step, start_at = _validate_fixed_step_chunker_args(chk_size, chk_step, start_at, stop_at)
#     if start_at is not None or stop_at is not None:
#         it = islice(it, start_at, stop_at)
#
#     iit = iter(it)
#     window = deque([], chk_size)
#     push_to_queue = window.extend
#
#     # and there after... push new step data to queue and yield queue contents
#     if not return_tail:
#         if chk_step == chk_size:
#             yield from zip(*([iit] * chk_step))
#         else:
#             chk = tuple(islice(iit, chk_size))
#             if len(chk) >= chk_size:
#                 yield chk
#                 push_to_queue(chk)
#                 for step_data in zip(*([iit] * chk_step)):
#                     push_to_queue(step_data)
#                     yield tuple(window)
#     else:
#         iiit = chain(iit, [nothing] * (chk_step - 1))
#         yield from map(strip_nothing_out, new_fixed_step_chunker(iiit, chk_size, chk_step, return_tail=False))
#         # step_nibs = zip(*([iiit] * chk_step))
#         # last_nib = next(step_nibs)
#         # for step_nib in step_nibs:
#         #     push_to_queue(last_nib)
#         #     yield tuple(window)
#         #     last_nib = step_nib
#         # push_to_queue(last_nib)
#         # last_part = list(window)
#         # yield strip_nothing_out(last_part)



from matplotlib.pylab import *
from numpy import *
from collections import Counter
import pandas as pd
from itertools import chain
from slang.core import DFLT_CHUNKER

# from agen.sound.audio import Sound
# from slang.featurizers import _rms_and_crossing_levels, _zero_crossing_ratio
# import seaborn as sns

# import i_do_not_exist.here
# import i_do_not_exist.there
# from i_also_do_not_exit import something



"""Slang utils"""

from collections import deque, defaultdict
from itertools import islice

import numpy as np
from numpy import array, nan, arange, unique, ones
from numpy.random import choice
from slang.util_data import displayable_unichr

from functools import partial
from contextlib import suppress

ModuleNotFoundIgnore = partial(
    suppress, ModuleNotFoundError
)  # just an alias for back-compatibility


def mk_callable(call_func):
    """A class decorator that adds a __call__ method. Specialized for sklearn models.

    >>> import numpy as np
    >>> from sklearn.decomposition import PCA
    >>>
    >>> CallablePCA = mk_callable('single_transform')(PCA)
    >>> pca = CallablePCA(n_components=3).fit(np.random.rand(100, 5))
    >>> x = np.random.rand(5)
    >>> all(pca(x) == pca.transform([x])[0])
    True
    >>>
    >>> from sklearn.neighbors import NearestNeighbors
    >>>
    >>> def nearest_neighbors_indices(self, x):
    ...     _, indices = self.kneighbors([x])
    ...     return indices[0]
    ...
    >>>
    >>> @mk_callable(nearest_neighbors_indices)
    ... class CallableKnn(NearestNeighbors):
    ...     '''NearestNeighbors with callable instances that give you the indices of the neighbors
    ...     without the kerfuffle.'''
    >>>
    >>> knn = CallableKnn().fit(np.arange(1000).reshape(200, 5))
    >>> x = np.array([10, 20, 30, 40, 50])  # say we have a single point we want to get neighbors for
    >>>
    >>> # This is the standard way to do it
    >>> _, indices = knn.kneighbors([x])
    >>> neighbors = indices[0]
    >>> neighbors
    array([6, 5, 7, 4, 8])
    >>> # but now, you can just do this instead:
    >>> knn(x)
    array([6, 5, 7, 4, 8])
    >>>
    >>> assert all(knn(x) == neighbors)
    """
    call_func_str = None
    if isinstance(call_func, str):
        call_func_str = call_func

    def add_call_method(cls):
        if call_func_str is not None:
            if hasattr(cls, call_func_str):
                cls.__call__ = getattr(cls, call_func_str)
            elif call_func_str.startswith('single_') and hasattr(
                cls, call_func_str[len('single_') :]
            ):
                call_method = getattr(cls, call_func_str[len('single_') :])

                def _call_func(self, x):
                    return call_method(self, [x])[0]

                cls.__call__ = _call_func
            else:
                raise ValueError(
                    f'call_func was specified by a string, but was neither the name of the method of the class'
                    f"nor 'single_METHOD_NAME' where METHOD_NAME is a method of the class: {call_func_str}"
                )
        else:
            assert callable(
                call_func
            ), f'call_func must be callable but was {call_func}'
            cls.__call__ = call_func
        return cls

    return add_call_method


####### Seeing snips ###################################################################################################

unichr_code_of_snip = array(
    displayable_unichr
    + list(unique(list(set(range(33, 20000)).difference(displayable_unichr))))
)
snip_of_unichr_code = (nan * ones(unichr_code_of_snip.max() + 1)).astype(int)
snip_of_unichr_code[unichr_code_of_snip] = arange(len(unichr_code_of_snip))

# assert that snip_of_unichr_code is the inverse of unichr_code_of_snip
assert all(snip_of_unichr_code[unichr_code_of_snip] == arange(len(unichr_code_of_snip)))


def snip_to_str(snip):
    return chr(unichr_code_of_snip[snip])


def snips_to_str(snips):
    return ''.join(map(chr, unichr_code_of_snip[list(snips)]))


def str_to_snips(snips_str):
    return array(snip_of_unichr_code[list(map(ord, snips_str))])


snips_of_str = str_to_snips  # alias for back-compatibility
str_of_snips = snips_to_str  # alias for back-compatibility

####### Misc ###########################################################################################################
ddir = lambda obj: [x for x in dir(obj) if not x.startswith('_')]


def row_euclidean_distance(A, B):
    """Euclidean distance between aligned rows of A. An array of length len(A) (==len(B)).

    >>> import numpy as np
    >>> A = np.arange(5 * 16).reshape((5, 16))
    >>> B = 1 + A

    >>> assert all(row_euclidean_distance(A, A) == np.zeros(5))
    >>> assert all(row_euclidean_distance(A, B) == np.array([4., 4., 4., 4., 4.]))

    Note: Not to be confused with the matrix of distances of all pairs of rows. Here, equivalent to the latter diagnonal (see below).

    ```
    from  sklearn.metrics.pairwise import euclidean_distances
    A = np.random.rand(5, 7)
    B = np.random.rand(5, 7)
    assert all(np.diag(euclidean_distances(A, B)) == row_euclidean_distance(A, B))
    ```

    """
    return np.sqrt(((A - B) ** 2).sum(axis=1))


def row_euclidean_distance_casting_to_array(A, B):
    return row_euclidean_distance(np.array(list(A)), np.array(list(B)))


class lazyprop:
    """
    A descriptor implementation of lazyprop (cached property) from David Beazley's "Python Cookbook" book.
    It's
    >>> class Test:
    ...     def __init__(self, a):
    ...         self.a = a
    ...     @lazyprop
    ...     def len(self):
    ...         print('generating "len"')
    ...         return len(self.a)
    >>> t = Test([0, 1, 2, 3, 4])
    >>> t.__dict__
    {'a': [0, 1, 2, 3, 4]}
    >>> t.len
    generating "len"
    5
    >>> t.__dict__
    {'a': [0, 1, 2, 3, 4], 'len': 5}
    >>> t.len
    5
    >>> # But careful when using lazyprop that no one will change the value of a without deleting the property first
    >>> t.a = [0, 1, 2]  # if we change a...
    >>> t.len  # ... we still get the old cached value of len
    5
    >>> del t.len  # if we delete the len prop
    >>> t.len  # ... then len being recomputed again
    generating "len"
    3
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            value = self.func(instance)
            setattr(instance, self.func.__name__, value)
            return value


def balanced_sample_maker(key_to_tag, max_n_keys_per_tag=7, random=False):
    """making a sample of the data (when you want to just test quickly)

    >>> mk_sample = balanced_sample_maker(key_to_tag=lambda k: k.split('/')[0],
    ...                                   max_n_keys_per_tag=2,
    ...                                   random=False)
    >>> mk_sample(['good/1', 'bad/1', 'good/2', 'good/3', 'good/4', 'bad/2', 'good/5', 'bad/3'])
    ['good/1', 'good/2', 'bad/1', 'bad/2']
    """

    def mk_balanced_sample(keys):
        sample = defaultdict(list)
        for k in keys:
            sample[key_to_tag(k)].append(k)
        counts = {k: len(v) for k, v in sample.items()}
        min_count = min(max_n_keys_per_tag, min(counts.values()))
        sample_keys = list()
        if random is False:
            for k in sample:
                sample_keys.extend(sample[k][:min_count])
        else:
            for k in sample:
                sample_keys.extend(choice(sample[k], size=min_count, replace=False))
        return sample_keys

    return mk_balanced_sample


import operator


def _apply_op(op, d1, dflt_1, d2, dflt_2):
    if isinstance(d2, dict):
        out = dict()
        for k, v1 in d1.items():
            v2 = d2.get(k, dflt_2)
            out[k] = op(v1, v2)
        for (
            k
        ) in (
            d2
        ):  # take care of the remainder (those keys in dict_2 that were not in dict_1)
            if k not in out:
                out[k] = op(dflt_1, d2[k])
    else:
        out = {k: op(v, d2) for k, v in d1.items()}
    return out


def _mk_op_method(op, dflt_1, dflt_2, for_reflexive_op=False):
    if not for_reflexive_op:

        def op_method(self, d):
            return self.__class__(_apply_op(op, self, dflt_1, d, dflt_2))

    else:

        def op_method(self, d):
            return self.__class__(_apply_op(op, d, dflt_1, self, dflt_2))

    return op_method


def _mk_unary_op_method(op):
    def op_method(self):
        return self.__class__({k: op(v) for k, v in self.items()})

    return op_method


# Notes: Not sure if mod has an identity.
# __and__ would have one theoretically (infinity 1s). didn't think about xor.
# __matmul__ has an identity DEPENDING ON square and if so, PER dimensions of matrix
# TODO: not sure of the appropriate defaults for __pow__.
_ops_and_identity = [
    ({'__add__', '__sub__', '__lshift__', '__rshift__', '__or__'}, 0),
    ({'__mul__', '__truediv__', '__floordiv__', '__pow__'}, 1),
    ({'__mod__', '__and__', '__xor__', '__matmul__'}, None),
]

_unary_ops = {'__pos__', '__neg__', '__abs__', '__invert__'}


class ArithmeDict(dict):
    """A dict, with arithmetic.
    A unary operator is just applied to all values.
    When a dict operates with a number, the operation is applied to each value of the dict.
    When a dict operates with another dict, the keys are aligned and the operation applied to the aligned values.

    The class is meant to be used in situations where pandas.Series would be used to operate with (sparse) vectors
    such as word counts, etc.

    Performance:

    In a nutshell, if you use pandas already in your app, then use pandas.Series instead.
    But, if you want weight packages (pandas isn't light), or have small dicts you want to operate on, use ArithmeDict.

    Note that both construction and operation are faster on ArithmeDict, for smaller sets.

    ```
    import pandas as pd

    t = ArithmeDict(a=1, b=2)
    tt = ArithmeDict(b=3, c=4)
    %timeit t + tt
    # 1.41 µs ± 41.6 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)

    ### versus ###

    t = pd.Series(dict(a=1, b=2))
    tt = pd.Series(dict(b=3, c=4))
    %timeit t + tt  # and not even what we want (see later)
    # 405 µs ± 7.65 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)
    % timeit pd.Series.add(t, tt, fill_value=0).to_dict()
    # 410 µs ± 11.9 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)

    ### but ###
    t = ArithmeDict({i: i for i in range(10000)})
    tt = ArithmeDict({i: i for i in range(5000, 15000)})
    %timeit t + tt
    # 3.22 ms ± 98.8 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)

    ### not so far from ###
    t = pd.Series({i: i for i in range(10000)})
    tt = pd.Series({i: i for i in range(5000, 15000)})
    %timeit pd.Series.add(t, tt, fill_value=0).to_dict()
    3.71 ms ± 100 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
    # and actually much slower than:
    %timeit pd.Series.add(t, tt, fill_value=0)
    575 µs ± 17.1 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)

    ```

    On the other hand, memory usage is inconclusive, because I don't know how to actually make the comparison.
    ```
    import pickle, sys, pandas

    t = ArithmeDict({i: i for i in range(10000)})
    sys.getsizeof(t), len(pickle.dumps(t))
    # (295032, 59539)

    t = pandas.Series({i: i for i in range(10000)})
    sys.getsizeof(t), len(pickle.dumps(t))
    # (160032, 240666)
    ```


    Notes for enhancement:

    When dict operates with/on a dict, and therefore we need to align keys,
    there are different merge and reduce options that may or may not make sense according to the value type and context.
    For example, should we really keep all keys and use operand defaults to get their values, or just drop
    those fields all together?
    Also, if we choose to keep all keys, what should the operand default be.
    Sometimes it might depend on the other operand (example matmul), or need to be created (example __concat__,
    since don't want the mutable list as a default), etc.

    >>> d1 = ArithmeDict(a=1, b=2)
    >>> d2 = ArithmeDict(b=3, c=4)
    >>>
    >>> # These are still dicts
    >>> isinstance(d1, dict)
    True
    >>> # and display as such
    >>> d1
    {'a': 1, 'b': 2}
    >>> d2
    {'b': 3, 'c': 4}
    >>>
    >>> # Unary operators (just applied to all values)
    >>> assert -d1 == {'a': -1, 'b': -2}
    >>> assert abs(-d1) == d1  # ... and in case that doesn't look impressive enough..
    >>> assert abs(ArithmeDict(a=-1, b=2, c=-3)) == {'a': 1, 'b': 2, 'c': 3}
    >>>
    >>> # An operation with a number is transferred to the values of the dict (applied to each).
    >>> assert d1 + 10 == {'a': 11, 'b': 12}
    >>> assert d1 - 10 == {'a': -9, 'b': -8}
    >>> assert d1 * 10 == {'a': 10, 'b': 20}
    >>> assert d1 / 10 == {'a': 0.1, 'b': 0.2}
    >>> assert d1 // 2 == {'a': 0, 'b': 1}
    >>> assert d1 ** 2 == {'a': 1, 'b': 4}
    >>> assert d2 % 2 == {'b': 1, 'c': 0}
    >>> assert d2 % 3 == {'b': 0, 'c': 1}
    >>> assert d2 >> 1 == {'b': 1, 'c': 2}  # shift all bits by one bit to the right
    >>> assert d2 << 1 == {'b': 6, 'c': 8}  # shift all bits by one bit to the left
    >>>
    >>> # An operation with another dict will align the keys and apply the operation to the aligned values.
    >>> assert d1 + d2 == {'a': 1, 'b': 5, 'c': 4}
    >>> assert d1 - d2 == {'a': 1, 'b': -1, 'c': -4}
    >>> assert d1 * d2 == {'a': 1, 'b': 6, 'c': 4}
    >>> assert d1 / d2 == {'a': 1, 'b': 0.6666666666666666, 'c': 0.25}
    >>> assert d2 // d1 == {'b': 1, 'c': 4, 'a': 1}
    >>> assert d1 ** d2 == {'a': 1, 'b': 8, 'c': 1}
    >>> assert ArithmeDict(a=10, b=10) % dict(a=3, b=4) == {'a': 1, 'b': 2}
    >>> assert d1 << d2 == {'a': 1, 'b': 16, 'c': 0}  # shifting bits
    >>> assert d1 + {'b': 3, 'c': 4} == {'a': 1, 'b': 5, 'c': 4}  # works when the right side is a normal dict
    >>> assert d1 + ArithmeDict() == d1
    >>> assert ArithmeDict() - d1 == -d1
    """

    for op in _unary_ops:
        locals()[op] = _mk_unary_op_method(getattr(operator, op))

    for ops, identity_val in _ops_and_identity:
        for op in ops:
            op_func = getattr(operator, op)
            locals()[op] = _mk_op_method(
                op_func,
                dflt_1=identity_val,
                dflt_2=identity_val,
                for_reflexive_op=False,
            )


def running_mean_gen(
    it, chk_size=2, chk_step=1
):  # TODO: A version of this with chk_step as well
    """
    Running mean (moving average) on iterator.
    Note: When input it is list-like, ut.stats.smooth.sliders version of running_mean is 4 times more efficient with
    big (but not too big, because happens in RAM) inputs.
    :param it: iterable
    :param chk_size: width of the window to take means from
    :return:

    >>> list(running_mean_gen([1, 3, 5, 7, 9], 2))
    [2.0, 4.0, 6.0, 8.0]
    >>> list(running_mean_gen([1, 3, 5, 7, 9], 2, chk_step=2))
    [2.0, 6.0]
    >>> list(running_mean_gen([1, 3, 5, 7, 9], 2, chk_step=3))
    [2.0, 8.0]
    >>> list(running_mean_gen([1, 3, 5, 7, 9], 3))
    [3.0, 5.0, 7.0]
    >>> list(running_mean_gen([1, -1, 1, -1], 2))
    [0.0, 0.0, 0.0]
    >>> list(running_mean_gen([-1, -2, -3, -4], 3))
    [-2.0, -3.0]
    """
    if chk_step is None:
        chk_step = chk_size
    if chk_step > 1:
        # TODO: perhaps there's a more efficient way. A way that would sum the values
        #  of every step and add them in bulk
        yield from islice(running_mean(it, chk_size), None, None, chk_step)
    else:
        it = iter(it)
        if chk_size > 1:

            c = 0
            fifo = deque([], maxlen=chk_size)
            for i, x in enumerate(it, 1):
                fifo.append(x)
                c += x
                if i >= chk_size:
                    break

            yield c / chk_size

            for x in it:
                c += x - fifo[0]  # NOTE: seems faster than fifo.popleft
                fifo.append(x)
                yield c / chk_size
        else:
            for x in it:
                yield x


def running_mean(arr, chk_size):
    c = np.cumsum(np.insert(arr, 0, [0]))
    return (c[chk_size:] - c[:-chk_size]) / chk_size



"""Various slang tools"""
from functools import singledispatch
from collections.abc import Sequence
from collections import Counter

DFLT_PRIOR_COUNT = 1


@singledispatch
def snip_probs(count_of_snip: dict, prior_count=DFLT_PRIOR_COUNT):
    total_count = sum(count_of_snip.values()) + len(count_of_snip) * prior_count
    return {
        snip: (count_of_snip + prior_count) / float(total_count)
        for snip, count_of_snip in count_of_snip.items()
    }


@snip_probs.register
def _(count_of_snip: Sequence, prior_count=DFLT_PRIOR_COUNT):
    return snip_probs(Counter(count_of_snip), prior_count=prior_count)
